from __future__ import annotations

import datetime as dt
import math
import re
from typing import cast

from relationalai.semantics.std.datetime_format import DateFormatField, DateFormatLiteral, parse_julia_datetime_format
from relationalai.semantics.backends.sql.dialect import (
    _DATE_INVALID_FORMAT_KINDS,
    _TZ_PATTERN_KINDS,
    _single_field_kind,
    _translate_duckdb_format,
    _translate_snowflake_format,
)

from .ir import (
    SQL2BinaryExpr,
    SQL2CastExpr,
    SQL2ColumnExpr,
    SQL2CreateViewOp,
    SQL2ExistsExpr,
    SQL2Expr,
    SQL2FromItem,
    SQL2FunctionExpr,
    SQL2Join,
    SQL2LiteralExpr,
    SQL2NullExpr,
    SQL2OrderByExpr,
    SQL2Query,
    SQL2RecursiveQuery,
    SQL2SelectItem,
    SQL2SelectQuery,
    SQL2SubquerySource,
    SQL2TableSource,
    SQL2UnionQuery,
    SQL2WindowExpr,
)


class SQL2Emitter:
    """Renderer for SQL2 structured IR."""

    class _IdentityDialect:
        @staticmethod
        def format_table_name(name: str) -> str:
            return name

    def __init__(self, *, dialect=None):
        self.dialect = dialect if dialect is not None else self._IdentityDialect()

    def emit(self, operation: SQL2CreateViewOp | SQL2Query) -> str:
        if isinstance(operation, SQL2CreateViewOp):
            return f"CREATE OR REPLACE VIEW {self._qid_path(operation.name)} AS\n{self.emit_query(operation.query)}"
        return self.emit_query(operation)

    def emit_query(self, query: SQL2Query) -> str:
        if isinstance(query, SQL2SelectQuery):
            return self._emit_select(query)
        if isinstance(query, SQL2UnionQuery):
            values_sql = self._emit_values_union(query)
            if values_sql is not None:
                return values_sql
            joiner = "\nUNION\n" if query.distinct else "\nUNION ALL\n"
            return joiner.join(self._wrap_query(part) for part in query.parts)
        if isinstance(query, SQL2RecursiveQuery):
            return self._emit_recursive(query)
        raise TypeError(f"Unsupported SQL2 query type: {type(query)!r}")

    def _emit_recursive(self, query: SQL2RecursiveQuery) -> str:
        if query.base_query is None or query.recursive_query is None:
            raise TypeError("SQL2RecursiveQuery requires both base_query and recursive_query.")
        if query.final_query is None:
            raise TypeError("SQL2RecursiveQuery requires final_query.")

        union_kw = "UNION" if query.distinct else "UNION ALL"
        column_sql = ""
        if query.column_names:
            column_sql = "(" + ", ".join(self._qid(name) for name in query.column_names) + ")"
        cte_sql = (
            f"WITH RECURSIVE {self._qid(query.cte_name)}{column_sql} AS (\n"
            f"{self._wrap_query(query.base_query)}\n"
            f"{union_kw}\n"
            f"{self._wrap_query(query.recursive_query)}\n"
            f")"
        )
        return f"{cte_sql}\n{self.emit_query(query.final_query)}"

    def _emit_select(self, query: SQL2SelectQuery) -> str:
        distinct = "DISTINCT " if query.distinct else ""
        select_items = (
            ",\n  ".join(self._emit_select_item(item) for item in query.select_items)
            or f'1 AS {self._qid("empty")}'
        )
        lines = [f"SELECT {distinct}\n  {select_items}"]
        limit_zero = self._is_canonical_empty_where(query.where)
        deferred_where: list[SQL2Expr] = []
        available_aliases: set[str] = set()
        if query.from_item is not None:
            lines.append(f"FROM {self._emit_from_item(query.from_item)}")
            available_aliases.add(query.from_item.alias)
        for join in query.joins:
            rendered_join = join
            if join.join_type.upper() == "INNER" and join.on:
                local_aliases = set(available_aliases)
                local_aliases.add(join.source.alias)
                local_on: list[SQL2Expr] = []
                for predicate in join.on:
                    aliases = self._expr_table_aliases(predicate)
                    if aliases.issubset(local_aliases):
                        local_on.append(predicate)
                    else:
                        deferred_where.append(predicate)
                rendered_join = SQL2Join(source=join.source, join_type=join.join_type, on=tuple(local_on))
            lines.append(self._emit_join(rendered_join))
            available_aliases.add(join.source.alias)
        effective_where = tuple((*query.where, *deferred_where))
        if effective_where and not limit_zero:
            where_sql = " AND ".join(self._emit_expr(predicate) for predicate in effective_where)
            lines.append(f"WHERE {where_sql}")
        if query.group_by:
            group_sql = ", ".join(self._emit_expr(expr) for expr in query.group_by)
            lines.append(f"GROUP BY {group_sql}")
        if query.having:
            having_sql = " AND ".join(self._emit_expr(predicate) for predicate in query.having)
            lines.append(f"HAVING {having_sql}")
        if limit_zero:
            lines.append("LIMIT 0")
        return "\n".join(lines)

    def _expr_table_aliases(self, expr: SQL2Expr) -> set[str]:
        if isinstance(expr, SQL2ColumnExpr):
            return {expr.table_alias}
        if isinstance(expr, SQL2BinaryExpr):
            return self._expr_table_aliases(expr.left) | self._expr_table_aliases(expr.right)
        if isinstance(expr, SQL2CastExpr):
            return self._expr_table_aliases(expr.expr)
        if isinstance(expr, SQL2FunctionExpr):
            out: set[str] = set()
            for arg in expr.args:
                out |= self._expr_table_aliases(arg)
            return out
        if isinstance(expr, SQL2WindowExpr):
            out: set[str] = set()
            for arg in expr.args:
                out |= self._expr_table_aliases(arg)
            for part in expr.partition_by:
                out |= self._expr_table_aliases(part)
            for order in expr.order_by:
                out |= self._expr_table_aliases(order.expr)
            return out
        return set()

    def _is_canonical_empty_where(self, where: tuple[SQL2Expr, ...]) -> bool:
        if len(where) != 1:
            return False
        predicate = where[0]
        return (
            isinstance(predicate, SQL2BinaryExpr)
            and predicate.op == "="
            and isinstance(predicate.left, SQL2LiteralExpr)
            and isinstance(predicate.right, SQL2LiteralExpr)
            and predicate.left.value == 1
            and predicate.right.value == 0
        )

    def _emit_values_union(self, query: SQL2UnionQuery) -> str | None:
        if query.distinct or not query.parts:
            return None

        aliases: tuple[str, ...] | None = None
        rows: list[tuple[SQL2Expr, ...]] = []
        for part in query.parts:
            if not isinstance(part, SQL2SelectQuery):
                return None
            if part.from_item is not None or part.joins or part.where or part.group_by or part.having or part.distinct:
                return None
            if not part.select_items or any(item.alias is None for item in part.select_items):
                return None

            part_aliases = tuple(item.alias for item in part.select_items)
            assert all(alias is not None for alias in part_aliases)
            if aliases is None:
                aliases = tuple(alias for alias in part_aliases if alias is not None)
            elif aliases != tuple(alias for alias in part_aliases if alias is not None):
                return None

            row_exprs = tuple(item.expr for item in part.select_items)
            if any(not isinstance(expr, (SQL2LiteralExpr, SQL2NullExpr)) for expr in row_exprs):
                return None
            rows.append(row_exprs)

        if aliases is None or not rows:
            return None

        values_rows = ",\n    ".join(
            "(" + ", ".join(self._emit_expr(expr) for expr in row) + ")"
            for row in rows
        )
        projected = ",\n  ".join(
            f'{self._qid("_values")}.{self._qid(alias)} AS {self._qid(alias)}'
            for alias in aliases
        )
        alias_columns = ", ".join(self._qid(alias) for alias in aliases)
        return (
            f"SELECT \n  {projected}\n"
            f"FROM (\n  VALUES\n    {values_rows}\n) AS {self._qid('_values')} ({alias_columns})"
        )

    def _emit_select_item(self, item: SQL2SelectItem) -> str:
        expr_sql = self._emit_expr(item.expr)
        if item.alias is None:
            return expr_sql
        return f"{expr_sql} AS {self._qid(item.alias)}"

    def _emit_join(self, join: SQL2Join) -> str:
        join_type = join.join_type.upper()
        join_sql = f"{join_type} JOIN {self._emit_from_item(join.source)}"
        if join.on:
            on_sql = " AND ".join(self._emit_expr(predicate) for predicate in join.on)
            return f"{join_sql} ON {on_sql}"
        if join_type == "INNER":
            return f"CROSS JOIN {self._emit_from_item(join.source)}"
        return f"{join_sql} ON TRUE"

    def _emit_from_item(self, item: SQL2FromItem) -> str:
        lateral = "LATERAL " if item.lateral else ""
        if isinstance(item.source, SQL2TableSource):
            source_sql = self._qid_path(item.source.name)
        elif isinstance(item.source, SQL2SubquerySource):
            source_sql = self._emit_subquery_source(item)
        else:
            raise TypeError(f"Unsupported SQL2 source: {type(item.source)!r}")
        return f"{lateral}{source_sql} AS {self._qid(item.alias)}"

    def _emit_subquery_source(self, item: SQL2FromItem) -> str:
        assert isinstance(item.source, SQL2SubquerySource)
        query = item.source.query
        source_sql = self._wrap_query(query)
        if isinstance(query, SQL2UnionQuery) and self._requires_from_union_wrapper(item):
            # Some dialects reject raw set-op forms in FROM/LATERAL positions.
            source_sql = f"(SELECT * FROM {source_sql} AS {self._qid('_union_src')})"
        return source_sql

    def _requires_from_union_wrapper(self, item: SQL2FromItem) -> bool:
        return False

    def _emit_expr(self, expr: SQL2Expr) -> str:
        if isinstance(expr, SQL2ColumnExpr):
            return f"{self._qid(expr.table_alias)}.{self._qid(expr.column_name)}"
        if isinstance(expr, SQL2LiteralExpr):
            if expr.value is None:
                return "NULL"
            if isinstance(expr.value, bool):
                return "TRUE" if expr.value else "FALSE"
            if isinstance(expr.value, float):
                if not math.isfinite(expr.value):
                    if math.isnan(expr.value):
                        return "NULL"
                    return "CAST('inf' AS DOUBLE)" if expr.value > 0 else "CAST('-inf' AS DOUBLE)"
            if isinstance(expr.value, dt.datetime):
                value = expr.value
                if value.tzinfo is not None:
                    value = value.astimezone(dt.timezone.utc).replace(tzinfo=None)
                literal = value.isoformat(sep=" ")
                escaped = literal.replace("'", "''")
                return f"CAST('{escaped}' AS TIMESTAMP)"
            if isinstance(expr.value, dt.date):
                return f"CAST('{expr.value.isoformat()}' AS DATE)"
            if isinstance(expr.value, str):
                return "'" + expr.value.replace("'", "''") + "'"
            return str(expr.value)
        if isinstance(expr, SQL2NullExpr):
            return "NULL"
        if isinstance(expr, SQL2FunctionExpr):
            return self._emit_function(expr)
        if isinstance(expr, SQL2WindowExpr):
            return self._emit_window(expr)
        if isinstance(expr, SQL2CastExpr):
            return f"CAST({self._emit_expr(expr.expr)} AS {expr.target_type})"
        if isinstance(expr, SQL2BinaryExpr):
            return f"({self._emit_expr(expr.left)} {expr.op} {self._emit_expr(expr.right)})"
        if isinstance(expr, SQL2ExistsExpr):
            prefix = "NOT EXISTS" if expr.negated else "EXISTS"
            return f"{prefix} {self._wrap_query(expr.query)}"
        raise TypeError(f"Unsupported SQL2 expr type: {type(expr)!r}")

    def _emit_function(self, expr: SQL2FunctionExpr) -> str:
        name = expr.name.lower()
        args_sql = [self._emit_expr(arg) for arg in expr.args]

        if (
            name == "count"
            and expr.distinct
            and len(expr.args) == 1
            and isinstance(expr.args[0], SQL2FunctionExpr)
            and expr.args[0].name.lower() == "row"
        ):
            rewritten = self._emit_count_distinct_row(expr.args[0].args)
            if rewritten is not None:
                return rewritten

        if name in {
            "day",
            "week",
            "month",
            "year",
            "hour",
            "minute",
            "second",
            "millisecond",
            "microsecond",
            "nanosecond",
        } and len(args_sql) == 1:
            return args_sql[0]
        if name in {"date_add", "date_subtract", "datetime_add", "datetime_subtract"} and len(expr.args) == 2:
            base_sql = args_sql[0]
            period = self._extract_period(expr.args[1])
            if period is not None:
                value_sql, unit = period
                return self._emit_date_add_or_subtract(
                    op_name=name,
                    base_sql=base_sql,
                    value_sql=value_sql,
                    unit=unit,
                )
        if name == "like" and len(args_sql) == 2:
            return f"({args_sql[0]} LIKE {args_sql[1]})"
        if name == "starts_with" and len(args_sql) == 2:
            return f"({args_sql[0]} LIKE CONCAT({args_sql[1]}, '%'))"
        if name == "ends_with" and len(args_sql) == 2:
            return f"({args_sql[0]} LIKE CONCAT('%', {args_sql[1]}))"
        if name == "contains" and len(args_sql) == 2:
            return f"({args_sql[0]} LIKE CONCAT('%', {args_sql[1]}, '%'))"
        if name == "regex_match" and len(args_sql) == 2:
            return self._emit_regex_match(pattern_sql=args_sql[0], value_sql=args_sql[1])
        if name == "md5" and len(args_sql) == 1:
            # CONCAT already yields a string expression; avoid redundant CAST noise.
            if isinstance(expr.args[0], SQL2FunctionExpr) and expr.args[0].name.lower() == "concat":
                return f"MD5({args_sql[0]})"
            return f"MD5(CAST({args_sql[0]} AS VARCHAR))"
        if name in {"strip", "trim"} and len(args_sql) == 1:
            return f"TRIM({args_sql[0]})"
        if name in {"natural_log", "ln"} and len(args_sql) == 1:
            return f"LN({args_sql[0]})"
        if name == "parse_float" and len(args_sql) == 1:
            return f"TRY_CAST({args_sql[0]} AS DOUBLE)"
        if name == "isinf" and len(args_sql) == 1:
            return f"COALESCE(ISINF({args_sql[0]}), FALSE)"
        if name == "isnan" and len(args_sql) == 1:
            return f"(({args_sql[0]}) IS NULL OR COALESCE(ISNAN({args_sql[0]}), FALSE))"
        if name in {"parse_number", "parse_decimal"} and len(args_sql) == 1:
            return f"TRY_CAST({args_sql[0]} AS DECIMAL(38,14))"
        if name in {"parse_number", "parse_decimal"} and len(expr.args) >= 3:
            precision_arg = expr.args[1]
            scale_arg = expr.args[2]
            if (
                isinstance(precision_arg, SQL2LiteralExpr)
                and isinstance(scale_arg, SQL2LiteralExpr)
                and isinstance(precision_arg.value, int)
                and isinstance(scale_arg.value, int)
            ):
                return f"TRY_CAST({args_sql[0]} AS DECIMAL({precision_arg.value},{scale_arg.value}))"
            return f"TRY_CAST({args_sql[0]} AS DECIMAL(38,14))"
        if name == "parse_int" and len(args_sql) == 1:
            return f"TRY_CAST({args_sql[0]} AS DECIMAL(38,0))"
        if name == "parse_uuid" and len(args_sql) == 1:
            return f"TRY_CAST({args_sql[0]} AS UUID)"
        if name == "uuid_to_string" and len(args_sql) == 1:
            return f"CAST({args_sql[0]} AS VARCHAR)"
        if name == "acot" and len(args_sql) == 1:
            return f"ATAN(1.0 / ({args_sql[0]}))"
        if name == "factorial" and len(args_sql) == 1:
            return f"FACTORIAL(CAST({args_sql[0]} AS INTEGER))"
        if name == "date_year" and len(args_sql) == 1:
            return f"EXTRACT(YEAR FROM {args_sql[0]})"
        if name == "date_month" and len(args_sql) == 1:
            return f"EXTRACT(MONTH FROM {args_sql[0]})"
        if name == "date_day" and len(args_sql) == 1:
            return f"EXTRACT(DAY FROM {args_sql[0]})"
        if name == "date_week" and len(args_sql) == 1:
            return f"WEEK({args_sql[0]})"
        if name == "date_quarter" and len(args_sql) == 1:
            return f"QUARTER({args_sql[0]})"
        if name == "date_dayofyear" and len(args_sql) == 1:
            return f"DAYOFYEAR({args_sql[0]})"
        if name in {"date_weekday", "date_isoweekday"} and len(args_sql) == 1:
            return self._emit_iso_weekday(args_sql[0])
        if name.startswith("datetime_") and len(args_sql) >= 1:
            base = args_sql[0]
            if len(args_sql) >= 2:
                base = self._convert_datetime_timezone_for_output(value_sql=base, tz_sql=args_sql[1])
            if name == "datetime_year":
                return f"EXTRACT(YEAR FROM {base})"
            if name == "datetime_month":
                return f"EXTRACT(MONTH FROM {base})"
            if name == "datetime_day":
                return f"EXTRACT(DAY FROM {base})"
            if name == "datetime_week":
                return f"WEEK({base})"
            if name == "datetime_quarter":
                return f"QUARTER({base})"
            if name == "datetime_dayofyear":
                return f"DAYOFYEAR({base})"
            if name in {"datetime_weekday", "datetime_isoweekday"}:
                return self._emit_iso_weekday(base)
            if name == "datetime_hour":
                return f"HOUR({base})"
            if name == "datetime_minute":
                return f"MINUTE({base})"
            if name == "datetime_second":
                return f"SECOND({base})"
        if name == "dates_period_days" and len(args_sql) == 2:
            return self._emit_dates_period_days(start_sql=args_sql[0], end_sql=args_sql[1])
        if name == "datetimes_period_milliseconds" and len(args_sql) == 2:
            return self._emit_datetimes_period_milliseconds(start_sql=args_sql[0], end_sql=args_sql[1])
        if name in {"date_diff", "datetime_diff"} and len(args_sql) == 3:
            return self._emit_date_diff_fn(unit_sql=args_sql[0], start_sql=args_sql[1], end_sql=args_sql[2])
        if name == "date_format" and len(args_sql) == 2:
            assert len(expr.args) == 2
            fmt_arg = expr.args[1]
            if isinstance(fmt_arg, SQL2LiteralExpr) and isinstance(fmt_arg.value, str):
                single_kind = _single_field_kind(fmt_arg.value)
                if single_kind in _DATE_INVALID_FORMAT_KINDS:
                    return "NULL"
                translated, _ = self._translate_datetime_format_literal(fmt_arg.value)
                fmt_sql = "'" + translated.replace("'", "''") + "'"
            else:
                fmt_sql = args_sql[1]
            return self._emit_date_format_call(value_sql=args_sql[0], format_sql=fmt_sql)
        if name == "datetime_format" and len(expr.args) >= 2:
            value_sql = args_sql[0]
            fmt_arg = expr.args[1]
            if len(expr.args) >= 3:
                tz_sql = args_sql[2]
                value_sql = self._convert_datetime_timezone_for_output(value_sql=value_sql, tz_sql=tz_sql)
            if isinstance(fmt_arg, SQL2LiteralExpr) and isinstance(fmt_arg.value, str):
                single_kind = _single_field_kind(fmt_arg.value)
                if single_kind == "tz_name":
                    return "'UTC'"
                if single_kind == "tz_offset":
                    return "'+00:00'"
                if single_kind == "millisecond":
                    return self._emit_datetime_millisecond(value_sql)
                translated, _ = self._translate_datetime_format_literal(fmt_arg.value)
                fmt_sql = "'" + translated.replace("'", "''") + "'"
            else:
                fmt_sql = args_sql[1]
            return self._emit_datetime_format_call(value_sql=value_sql, format_sql=fmt_sql)
        if name == "parse_date" and len(expr.args) == 2:
            fmt_arg = expr.args[1]
            if isinstance(fmt_arg, SQL2LiteralExpr) and isinstance(fmt_arg.value, str):
                translated, _ = self._translate_datetime_format_literal(fmt_arg.value)
                fmt_sql = "'" + translated.replace("'", "''") + "'"
                return self._emit_parse_date(value_sql=args_sql[0], format_sql=fmt_sql)
        if name == "parse_datetime" and len(expr.args) == 2:
            fmt_arg = expr.args[1]
            if isinstance(fmt_arg, SQL2LiteralExpr) and isinstance(fmt_arg.value, str):
                translated, tokens = self._translate_datetime_format_literal(fmt_arg.value)
                fmt_sql = "'" + translated.replace("'", "''") + "'"
                return self._emit_parse_datetime(value_sql=args_sql[0], format_sql=fmt_sql, tokens=tokens)
        if name == "construct_datetime_ms_tz" and len(args_sql) == 8:
            return self._emit_construct_datetime_ms_tz(args_sql=args_sql)
        if name == "construct_date" and len(args_sql) == 3:
            return self._emit_construct_date(args_sql=args_sql)
        if name == "construct_date_from_datetime" and len(args_sql) == 2:
            return self._emit_construct_date_from_datetime(value_sql=args_sql[0], tz_sql=args_sql[1])
        if name == "substring" and len(args_sql) in {2, 3}:
            start = f"CAST({args_sql[1]} AS BIGINT)"
            if len(args_sql) == 2:
                return f"SUBSTRING({args_sql[0]}, ({start} + 1))"
            stop = f"CAST({args_sql[2]} AS BIGINT)"
            return f"SUBSTRING({args_sql[0]}, ({start} + 1), ({stop} - {start}))"
        if name in {"split_part", "split", "str_split"} and len(args_sql) == 3:
            index_sql = f"(CAST({args_sql[2]} AS BIGINT) + 1)"
            return f"SPLIT_PART({args_sql[0]}, {args_sql[1]}, {index_sql})"

        rendered_args = ", ".join(args_sql)
        if expr.distinct and rendered_args:
            rendered_args = f"DISTINCT {rendered_args}"
        return f"{expr.name.upper()}({rendered_args})"

    def _translate_datetime_format_literal(self, fmt: str) -> tuple[str, set[str]]:
        return _translate_duckdb_format(fmt)

    def _emit_date_add_or_subtract(
        self,
        *,
        op_name: str,
        base_sql: str,
        value_sql: str,
        unit: str,
    ) -> str:
        sign = "+" if op_name in {"date_add", "datetime_add"} else "-"
        if unit == "NANOSECOND":
            return f"({base_sql} {sign} (({value_sql}) * INTERVAL 1 MICROSECOND) / 1000)"
        if unit == "MILLISECOND":
            signed_value = f"({value_sql})" if sign == "+" else f"(-({value_sql}))"
            day_part = f"CAST(({signed_value}) / 86400000 AS BIGINT)"
            ms_part = f"CAST(({signed_value}) % 86400000 AS BIGINT)"
            return (
                f"({base_sql} + ({day_part}) * INTERVAL 1 DAY "
                f"+ ({ms_part}) * INTERVAL 1 MILLISECOND)"
            )
        return f"({base_sql} {sign} ({value_sql}) * INTERVAL 1 {unit})"

    def _emit_regex_match(self, *, pattern_sql: str, value_sql: str) -> str:
        return f"REGEXP_MATCHES({value_sql}, {pattern_sql})"

    def _emit_iso_weekday(self, value_sql: str) -> str:
        return f"DATE_PART('isodow', {value_sql})"

    def _convert_datetime_timezone_for_output(self, *, value_sql: str, tz_sql: str) -> str:
        return f"TIMEZONE({tz_sql}, TIMEZONE('UTC', {value_sql}))"

    def _emit_dates_period_days(self, *, start_sql: str, end_sql: str) -> str:
        return f"DATE_DIFF('day', {start_sql}, {end_sql})"

    def _emit_datetimes_period_milliseconds(self, *, start_sql: str, end_sql: str) -> str:
        return f"DATE_DIFF('millisecond', {start_sql}, {end_sql})"

    def _emit_date_diff_fn(self, *, unit_sql: str, start_sql: str, end_sql: str) -> str:
        return f"DATE_DIFF({unit_sql}, {start_sql}, {end_sql})"

    def _emit_date_format_call(self, *, value_sql: str, format_sql: str) -> str:
        return f"STRFTIME({value_sql}, {format_sql})"

    def _emit_datetime_format_call(self, *, value_sql: str, format_sql: str) -> str:
        return f"STRFTIME({value_sql}, {format_sql})"

    def _emit_datetime_millisecond(self, value_sql: str) -> str:
        return f"CAST((CAST(DATE_PART('milliseconds', {value_sql}) AS BIGINT) % 1000) AS VARCHAR)"

    def _emit_parse_date(self, *, value_sql: str, format_sql: str) -> str:
        return f"CAST(STRPTIME({value_sql}, {format_sql}) AS DATE)"

    def _emit_parse_datetime(self, *, value_sql: str, format_sql: str, tokens: set[str]) -> str:
        parsed = f"STRPTIME({value_sql}, {format_sql})"
        if tokens & _TZ_PATTERN_KINDS:
            return f"CAST(TIMEZONE('UTC', {parsed}) AS TIMESTAMP)"
        return f"CAST({parsed} AS TIMESTAMP)"

    def _emit_construct_datetime_ms_tz(self, *, args_sql: list[str]) -> str:
        sec_with_fraction = f"(CAST({args_sql[5]} AS DOUBLE) + CAST({args_sql[6]} AS DOUBLE) / 1000.0)"
        local_ts = (
            "MAKE_TIMESTAMP("
            f"CAST({args_sql[0]} AS BIGINT), "
            f"CAST({args_sql[1]} AS BIGINT), "
            f"CAST({args_sql[2]} AS BIGINT), "
            f"CAST({args_sql[3]} AS BIGINT), "
            f"CAST({args_sql[4]} AS BIGINT), "
            f"{sec_with_fraction})"
        )
        return f"CAST(TIMEZONE('UTC', TIMEZONE({args_sql[7]}, {local_ts})) AS TIMESTAMP)"

    def _emit_construct_date(self, *, args_sql: list[str]) -> str:
        return (
            "MAKE_DATE("
            f"CAST({args_sql[0]} AS BIGINT), "
            f"CAST({args_sql[1]} AS BIGINT), "
            f"CAST({args_sql[2]} AS BIGINT))"
        )

    def _emit_construct_date_from_datetime(self, *, value_sql: str, tz_sql: str) -> str:
        return f"CAST(TIMEZONE({tz_sql}, TIMEZONE('UTC', {value_sql})) AS DATE)"

    def _emit_count_distinct_row(self, row_args: tuple[SQL2Expr, ...]) -> str | None:
        return None

    def _emit_window(self, expr: SQL2WindowExpr) -> str:
        args_sql = ", ".join(self._emit_expr(arg) for arg in expr.args)
        if expr.distinct and args_sql:
            args_sql = f"DISTINCT {args_sql}"
        fn_sql = f"{expr.name.upper()}({args_sql})"

        parts: list[str] = []
        if expr.partition_by:
            parts.append("PARTITION BY " + ", ".join(self._emit_expr(part) for part in expr.partition_by))
        if expr.order_by:
            parts.append(
                "ORDER BY "
                + ", ".join(
                    self._emit_order_item(item)
                    for item in expr.order_by
                )
            )
        over_sql = " ".join(parts)
        if over_sql:
            return f"{fn_sql} OVER ({over_sql})"
        return f"{fn_sql} OVER ()"

    def _emit_order_item(self, item: SQL2OrderByExpr) -> str:
        direction = "DESC" if item.descending else "ASC"
        return f"{self._emit_expr(item.expr)} {direction}"

    def _extract_period(self, period_expr: SQL2Expr) -> tuple[str, str] | None:
        # Period constructor outputs can be type-coerced (for example,
        # CAST(WEEK(x) AS DECIMAL(...))). Strip casts so we preserve units.
        while isinstance(period_expr, SQL2CastExpr):
            period_expr = period_expr.expr
        if not isinstance(period_expr, SQL2FunctionExpr) or len(period_expr.args) != 1:
            return None
        unit = period_expr.name.upper()
        allowed_units = {
            "NANOSECOND",
            "MICROSECOND",
            "MILLISECOND",
            "SECOND",
            "MINUTE",
            "HOUR",
            "DAY",
            "WEEK",
            "MONTH",
            "YEAR",
        }
        if unit not in allowed_units:
            return None
        return self._emit_expr(period_expr.args[0]), unit

    def _wrap_query(self, query: SQL2Query) -> str:
        return f"(\n{self.emit_query(query)}\n)"

    def _qid(self, name: str) -> str:
        escaped = name.replace('"', '""')
        return f'"{escaped}"'

    def _qid_path(self, path: str) -> str:
        return ".".join(self._qid(part) for part in path.split("."))

    def _quote_identifier(self, name: str) -> str:
        return self._qid(name)


class SQL2SnowflakeEmitter(SQL2Emitter):
    """Snowflake-targeted SQL2 emitter."""

    def emit(self, operation: SQL2CreateViewOp | SQL2Query) -> str:
        if isinstance(operation, SQL2CreateViewOp):
            # Snowflake can produce incorrect results on some self-join + inequality
            # patterns when relation definitions remain deeply nested views. Materialize
            # install-time relations as tables and keep query-time SQL mechanical.
            return f"CREATE OR REPLACE TABLE {self._qid_path(operation.name)} AS\n{self.emit_query(operation.query)}"
        return self.emit_query(operation)

    def _emit_select(self, query: SQL2SelectQuery) -> str:
        range_sql = self._emit_range_unnest_select(query)
        if range_sql is not None:
            return range_sql
        return super()._emit_select(query)

    def _emit_range_unnest_select(self, query: SQL2SelectQuery) -> str | None:
        # Snowflake has no UNNEST(RANGE(...)) scalar form. Lowering emits this
        # shape for std range/split helpers, so rewrite it to FLATTEN over
        # ARRAY_GENERATE_RANGE.
        if (
            query.from_item is not None
            or query.joins
            or query.where
            or query.group_by
            or query.having
            or query.distinct
            or len(query.select_items) != 1
        ):
            return None
        item = query.select_items[0]
        expr = item.expr
        if not isinstance(expr, SQL2FunctionExpr) or expr.name.lower() != "unnest" or len(expr.args) != 1:
            return None
        range_expr = expr.args[0]
        if not isinstance(range_expr, SQL2FunctionExpr) or range_expr.name.lower() != "range" or len(range_expr.args) != 3:
            return None

        start_sql = self._emit_expr(range_expr.args[0])
        stop_sql = self._emit_expr(range_expr.args[1])
        step_sql = self._emit_expr(range_expr.args[2])
        alias = item.alias or "value"
        return (
            "SELECT\n"
            f"  CAST(\"_f\".\"VALUE\" AS BIGINT) AS {self._qid(alias)}\n"
            "FROM TABLE(FLATTEN(INPUT => ARRAY_GENERATE_RANGE("
            f"CAST({start_sql} AS BIGINT), "
            f"CAST({stop_sql} AS BIGINT), "
            f"CAST({step_sql} AS BIGINT)"
            "))) AS \"_f\""
        )

    def _emit_function(self, expr: SQL2FunctionExpr) -> str:
        name = expr.name.lower()
        args_sql = [self._emit_expr(arg) for arg in expr.args]

        if name == "levenshtein" and len(args_sql) == 2:
            return f"EDITDISTANCE({args_sql[0]}, {args_sql[1]})"
        if name in {"string_split", "str_split"} and len(args_sql) == 2:
            return f"SPLIT({args_sql[0]}, {args_sql[1]})"
        if name == "array_length" and len(args_sql) == 1:
            return f"ARRAY_SIZE({args_sql[0]})"
        if name == "isinf" and len(args_sql) == 1:
            return (
                "COALESCE(UPPER(TO_VARCHAR("
                f"{args_sql[0]}"
                ")) IN ('INF', '-INF', 'INFINITY', '-INFINITY'), FALSE)"
            )
        if name == "isnan" and len(args_sql) == 1:
            return (
                f"(({args_sql[0]}) IS NULL OR "
                "COALESCE(UPPER(TO_VARCHAR("
                f"{args_sql[0]}"
                ")) = 'NAN', FALSE))"
            )
        if name == "acot" and len(args_sql) == 1:
            return (
                f"CASE WHEN ({args_sql[0]}) = 0 THEN PI() / 2 "
                f"ELSE ATAN(1 / TO_DOUBLE({args_sql[0]})) END"
            )
        if name == "date_format" and len(expr.args) == 2:
            fmt_arg = expr.args[1]
            if isinstance(fmt_arg, SQL2LiteralExpr) and isinstance(fmt_arg.value, str):
                single_kind = _single_field_kind(fmt_arg.value)
                if single_kind in _DATE_INVALID_FORMAT_KINDS:
                    return "NULL"
                if self._snowflake_format_requires_manual(fmt_arg.value):
                    return self._emit_snowflake_manual_format(value_sql=args_sql[0], fmt=fmt_arg.value)
        if name == "datetime_format" and len(expr.args) >= 2:
            value_sql = args_sql[0]
            fmt_arg = expr.args[1]
            if len(expr.args) >= 3:
                tz_sql = args_sql[2]
                value_sql = self._convert_datetime_timezone_for_output(value_sql=value_sql, tz_sql=tz_sql)
            if isinstance(fmt_arg, SQL2LiteralExpr) and isinstance(fmt_arg.value, str):
                single_kind = _single_field_kind(fmt_arg.value)
                if single_kind == "tz_name":
                    return "'UTC'"
                if single_kind == "tz_offset":
                    return "'+00:00'"
                if single_kind == "millisecond":
                    return self._emit_datetime_millisecond(value_sql)
                if self._snowflake_format_requires_manual(fmt_arg.value):
                    return self._emit_snowflake_manual_format(value_sql=value_sql, fmt=fmt_arg.value)

        return super()._emit_function(expr)

    def _snowflake_format_requires_manual(self, fmt: str) -> bool:
        for part in parse_julia_datetime_format(fmt):
            if not isinstance(part, DateFormatField):
                continue
            if part.kind.endswith("_unpadded") or part.kind in {"weekday_full", "month_full"}:
                return True
        return False

    def _emit_snowflake_manual_format(self, *, value_sql: str, fmt: str) -> str:
        part_sql: list[str] = []
        token_sql_by_kind = {
            "weekday_abbr": "DY",
            "month_abbr": "MON",
            "tz_offset": "TZHTZM",
            "tz_name": "TZD",
            "ampm": "AM",
            "year_2": "YY",
        }
        for part in parse_julia_datetime_format(fmt):
            if isinstance(part, DateFormatLiteral):
                text = part.text.replace("'", "''")
                part_sql.append(f"'{text}'")
                continue
            assert isinstance(part, DateFormatField)
            kind = part.kind
            if kind == "year_4":
                part_sql.append(f"TO_VARCHAR(EXTRACT(YEAR FROM {value_sql}))")
            elif kind == "year_2":
                part_sql.append(f"LPAD(TO_VARCHAR(MOD(EXTRACT(YEAR FROM {value_sql}), 100)), 2, '0')")
            elif kind == "month_unpadded":
                part_sql.append(f"TO_VARCHAR(EXTRACT(MONTH FROM {value_sql}))")
            elif kind == "month_padded":
                part_sql.append(f"LPAD(TO_VARCHAR(EXTRACT(MONTH FROM {value_sql})), 2, '0')")
            elif kind == "day_unpadded":
                part_sql.append(f"TO_VARCHAR(EXTRACT(DAY FROM {value_sql}))")
            elif kind == "day_padded":
                part_sql.append(f"LPAD(TO_VARCHAR(EXTRACT(DAY FROM {value_sql})), 2, '0')")
            elif kind == "hour24_unpadded":
                part_sql.append(f"TO_VARCHAR(EXTRACT(HOUR FROM {value_sql}))")
            elif kind == "hour24_padded":
                part_sql.append(f"LPAD(TO_VARCHAR(EXTRACT(HOUR FROM {value_sql})), 2, '0')")
            elif kind == "minute_unpadded":
                part_sql.append(f"TO_VARCHAR(EXTRACT(MINUTE FROM {value_sql}))")
            elif kind == "minute_padded":
                part_sql.append(f"LPAD(TO_VARCHAR(EXTRACT(MINUTE FROM {value_sql})), 2, '0')")
            elif kind == "second_unpadded":
                part_sql.append(f"TO_VARCHAR(EXTRACT(SECOND FROM {value_sql}))")
            elif kind == "second_padded":
                part_sql.append(f"LPAD(TO_VARCHAR(EXTRACT(SECOND FROM {value_sql})), 2, '0')")
            elif kind == "hour12_unpadded":
                part_sql.append(
                    f"TO_VARCHAR(IFF(MOD(EXTRACT(HOUR FROM {value_sql}), 12) = 0, 12, MOD(EXTRACT(HOUR FROM {value_sql}), 12)))"
                )
            elif kind == "hour12_padded":
                part_sql.append(
                    f"LPAD(TO_VARCHAR(IFF(MOD(EXTRACT(HOUR FROM {value_sql}), 12) = 0, 12, MOD(EXTRACT(HOUR FROM {value_sql}), 12))), 2, '0')"
                )
            elif kind == "weekday_full":
                part_sql.append(
                    f"CASE DAYOFWEEKISO({value_sql}) "
                    "WHEN 1 THEN 'Monday' WHEN 2 THEN 'Tuesday' WHEN 3 THEN 'Wednesday' "
                    "WHEN 4 THEN 'Thursday' WHEN 5 THEN 'Friday' WHEN 6 THEN 'Saturday' "
                    "ELSE 'Sunday' END"
                )
            elif kind == "month_full":
                part_sql.append(
                    f"CASE EXTRACT(MONTH FROM {value_sql}) "
                    "WHEN 1 THEN 'January' WHEN 2 THEN 'February' WHEN 3 THEN 'March' "
                    "WHEN 4 THEN 'April' WHEN 5 THEN 'May' WHEN 6 THEN 'June' "
                    "WHEN 7 THEN 'July' WHEN 8 THEN 'August' WHEN 9 THEN 'September' "
                    "WHEN 10 THEN 'October' WHEN 11 THEN 'November' ELSE 'December' END"
                )
            else:
                token = token_sql_by_kind.get(kind)
                if token is None:
                    translated, _ = self._translate_datetime_format_literal(fmt)
                    escaped = translated.replace("'", "''")
                    return f"TO_VARCHAR({value_sql}, '{escaped}')"
                part_sql.append(f"TO_VARCHAR({value_sql}, '{token}')")
        if not part_sql:
            return "''"
        return " || ".join(part_sql)

    def _requires_from_union_wrapper(self, item: SQL2FromItem) -> bool:
        return item.lateral

    def _qid_path(self, path: str) -> str:
        return ".".join(self._snowflake_path_identifier(part) for part in path.split("."))

    def _snowflake_path_identifier(self, name: str) -> str:
        token = name.strip()
        if token == "*":
            return token
        quote_identifier = getattr(self.dialect, "quote_identifier", None)
        if callable(quote_identifier):
            return cast(str, quote_identifier(token))
        if token.startswith('"') and token.endswith('"'):
            return token
        if re.fullmatch(r"[A-Za-z_][A-Za-z0-9_$]*", token):
            return token.upper()
        return self._qid(token)

    def _translate_datetime_format_literal(self, fmt: str) -> tuple[str, set[str]]:
        return _translate_snowflake_format(fmt)

    def _emit_date_add_or_subtract(
        self,
        *,
        op_name: str,
        base_sql: str,
        value_sql: str,
        unit: str,
    ) -> str:
        sign = 1 if op_name in {"date_add", "datetime_add"} else -1
        return f"DATEADD({unit.lower()}, ({sign} * ({value_sql})), {base_sql})"

    def _emit_regex_match(self, *, pattern_sql: str, value_sql: str) -> str:
        return f"REGEXP_LIKE({value_sql}, {pattern_sql})"

    def _emit_iso_weekday(self, value_sql: str) -> str:
        return f"DAYOFWEEKISO({value_sql})"

    def _convert_datetime_timezone_for_output(self, *, value_sql: str, tz_sql: str) -> str:
        return f"CONVERT_TIMEZONE('UTC', {tz_sql}, {value_sql})"

    def _emit_dates_period_days(self, *, start_sql: str, end_sql: str) -> str:
        return f"DATEDIFF(day, {start_sql}, {end_sql})"

    def _emit_datetimes_period_milliseconds(self, *, start_sql: str, end_sql: str) -> str:
        return f"DATEDIFF(millisecond, {start_sql}, {end_sql})"

    def _emit_date_diff_fn(self, *, unit_sql: str, start_sql: str, end_sql: str) -> str:
        return f"DATEDIFF({unit_sql}, {start_sql}, {end_sql})"

    def _emit_date_format_call(self, *, value_sql: str, format_sql: str) -> str:
        return f"TO_VARCHAR({value_sql}, {format_sql})"

    def _emit_datetime_format_call(self, *, value_sql: str, format_sql: str) -> str:
        return f"TO_VARCHAR({value_sql}, {format_sql})"

    def _emit_datetime_millisecond(self, value_sql: str) -> str:
        frac = f"TO_VARCHAR({value_sql}, 'FF3')"
        return f"IFF({frac} = '000', '0', LTRIM({frac}, '0'))"

    def _emit_parse_date(self, *, value_sql: str, format_sql: str) -> str:
        return f"TRY_TO_DATE({value_sql}, {format_sql})"

    def _emit_parse_datetime(self, *, value_sql: str, format_sql: str, tokens: set[str]) -> str:
        if tokens & _TZ_PATTERN_KINDS:
            return f"CAST(CONVERT_TIMEZONE('UTC', TRY_TO_TIMESTAMP_TZ({value_sql}, {format_sql})) AS TIMESTAMP_NTZ)"
        return f"CAST(TRY_TO_TIMESTAMP_NTZ({value_sql}, {format_sql}) AS TIMESTAMP_NTZ)"

    def _emit_construct_datetime_ms_tz(self, *, args_sql: list[str]) -> str:
        local_ts = (
            "TO_TIMESTAMP_NTZ("
            f"LPAD(CAST({args_sql[0]} AS VARCHAR), 4, '0') || '-' || "
            f"LPAD(CAST({args_sql[1]} AS VARCHAR), 2, '0') || '-' || "
            f"LPAD(CAST({args_sql[2]} AS VARCHAR), 2, '0') || ' ' || "
            f"LPAD(CAST({args_sql[3]} AS VARCHAR), 2, '0') || ':' || "
            f"LPAD(CAST({args_sql[4]} AS VARCHAR), 2, '0') || ':' || "
            f"LPAD(CAST({args_sql[5]} AS VARCHAR), 2, '0') || '.' || "
            f"LPAD(CAST({args_sql[6]} AS VARCHAR), 3, '0'), "
            "'YYYY-MM-DD HH24:MI:SS.FF3')"
        )
        return f"CAST(CONVERT_TIMEZONE({args_sql[7]}, 'UTC', {local_ts}) AS TIMESTAMP_NTZ)"

    def _emit_construct_date(self, *, args_sql: list[str]) -> str:
        return (
            "DATE_FROM_PARTS("
            f"CAST({args_sql[0]} AS BIGINT), "
            f"CAST({args_sql[1]} AS BIGINT), "
            f"CAST({args_sql[2]} AS BIGINT))"
        )

    def _emit_construct_date_from_datetime(self, *, value_sql: str, tz_sql: str) -> str:
        return f"CAST(CONVERT_TIMEZONE('UTC', {tz_sql}, {value_sql}) AS DATE)"

    def _emit_count_distinct_row(self, row_args: tuple[SQL2Expr, ...]) -> str | None:
        if not row_args:
            return "COUNT(*)"
        rendered_args = ", ".join(self._emit_expr(arg) for arg in row_args)
        return f"COUNT(DISTINCT {rendered_args})"



"""Common configuration sections for :func:`relationalai.config.Config`.

Most projects configure PyRel using a YAML file (typically ``raiconfig.yaml``).
This module defines the Pydantic models that validate the nested sections in
that file (for example ``execution``, ``data``, and ``debug``).

You can also configure everything in code by passing dictionaries to
the :func:`~relationalai.config.create_config` factory function (useful for programmatic
setup or for overrides in tests).

For reasoner (engine) configuration, see :mod:`relationalai.config.config_reasoners_fields`.
"""

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field, field_validator

from ..util.docutils import include_in_docs

MAX_DATA_FRESHNESS_MINS = 30240  # 3 weeks, as we might start cleaning up data after that

# include this module in documentation
__include_in_docs__ = True


@include_in_docs
class ExecutionConfig(BaseModel):
    """Configure execution middleware behavior.

    This controls client-side middleware features such as metrics/logging
    and (optionally) automatic retries with exponential backoff.

    Attributes
    ----------
    metrics : bool
        Enable execution metrics collection.
    logging : bool
        Enable structured execution logging.
    retries : ExecutionConfig.RetriesConfig
        Retry and backoff policy.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    execution:
        metrics: true
        retries:
            enabled: true
            max_attempts: 5
            base_delay_s: 0.5
            max_delay_s: 10.0
    ```

    Configure execution settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(
    ...     execution={
    ...         "metrics": True,
    ...         "retries": {"enabled": True, "max_attempts": 5, "base_delay_s": 0.5, "max_delay_s": 10.0},
    ...     },
    ... )

    Configure execution settings using an explicit ExecutionConfig instance:

    >>> from relationalai.config import create_config, ExecutionConfig
    >>> cfg = create_config(
    ...     execution=ExecutionConfig(
    ...         metrics=True,
    ...         retries=ExecutionConfig.RetriesConfig(enabled=True, max_attempts=5, base_delay_s=0.5, max_delay_s=10.0),
    ...     ),
    ... )
    """

    metrics: bool = Field(default=False, description="Enable execution metrics collection (via middleware)")
    logging: bool = Field(default=False, description="Enable execution structured logging (via middleware)")

    @include_in_docs
    class RetriesConfig(BaseModel):
        """Configure execution retries and exponential backoff.

        This model is used for the :attr:`~relationalai.config.ExecutionConfig.retries`
        nested section in :class:`~relationalai.config.ExecutionConfig` and
        controls whether retries are enabled and how delay grows between attempts.

        Attributes
        ----------
        enabled : bool
            Whether to retry failed requests.
        max_attempts : int
            Maximum number of attempts (including the first try).
        base_delay_s : float
            Base backoff delay in seconds.
        max_delay_s : float
            Maximum backoff delay in seconds.
        jitter : float
            Jitter factor applied to each delay (for example, ``0.2`` means ±20%).
        """

        enabled: bool = Field(default=False, description="Enable execution retries/backoff (via middleware)")
        max_attempts: int = Field(default=3, description="Max attempts for retries (when enabled)")
        base_delay_s: float = Field(default=0.25, description="Base delay for exponential backoff (seconds)")
        max_delay_s: float = Field(default=5.0, description="Max backoff delay (seconds)")
        jitter: float = Field(default=0.2, description="Jitter factor for backoff (0.2 => +/-20%)")

    retries: RetriesConfig = Field(default_factory=RetriesConfig)


@include_in_docs
class DataConfig(BaseModel):
    """Configure data loading, streaming, and export behavior.

    This controls client-side defaults used by helpers that load data,
    wait for streams to synchronize, and download exported results.

    Attributes
    ----------
    wait_for_stream_sync : bool
        Wait for stream synchronization before processing.
    ensure_change_tracking : bool
        Enable change tracking for data modifications.
    data_freshness_mins : int, optional
        Data freshness timeout in minutes.
    query_timeout_mins : int, optional
        Query timeout in minutes.
    download_url_type : {"internal", "external"}, optional
        Type of download URL used for data exports.
    check_column_types : bool
        Check column types during data loading.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    data:
        wait_for_stream_sync: false
        download_url_type: external
    ```

    Configure data settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(
    ...     data={"wait_for_stream_sync": False, "download_url_type": "external"},
    ... )

    Configure data settings using an explicit DataConfig instance:

    >>> from relationalai.config import create_config, DataConfig
    >>> cfg = create_config(
    ...     data=DataConfig(wait_for_stream_sync=False, download_url_type="external"),
    ... )
    """

    wait_for_stream_sync: bool = Field(default=True, description="Wait for stream synchronization before processing")
    ensure_change_tracking: bool = Field(default=False, description="Enable change tracking for data modifications")
    data_freshness_mins: int | None = Field(default=None, description="Data freshness timeout in minutes")

    @field_validator("data_freshness_mins")
    @classmethod
    def _clamp_data_freshness(cls, v: int | None) -> int | None:
        if v is not None and v > MAX_DATA_FRESHNESS_MINS:
            from relationalai.util.error import warn
            warn(
                "InvalidConfigValue",
                f"data_freshness_mins={v} exceeds the maximum of {MAX_DATA_FRESHNESS_MINS} "
                f"(3 weeks) and will be clamped to {MAX_DATA_FRESHNESS_MINS}.",
            )
            return MAX_DATA_FRESHNESS_MINS
        return v

    query_timeout_mins: int | None = Field(
        default=None, description="Query timeout in minutes - aborts queries that exceed this duration"
    )
    download_url_type: Literal["internal", "external"] | None = Field(
        default=None, description="Type of download URL for data exports (internal or external)"
    )
    check_column_types: bool = Field(default=True, description="Check column types during data loading")


@include_in_docs
class CompilerConfig(BaseModel):
    """Configure compiler behavior for query translation.

    This controls client-side compilation defaults such as strictness,
    output formatting, and a few experimental optimization flags.

    Attributes
    ----------
    dry_run : bool
        Run compilation in dry-run mode.
    strict : bool
        Enable strict validation mode.
    soft_type_errors : bool
        Treat type errors as warnings instead of failures.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    compiler:
        strict: true
        soft_type_errors: true
    ```

    Configure compiler settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(
    ...     compiler={"dry_run": True, "strict": True, "soft_type_errors": True},
    ... )

    Configure compiler settings using an explicit CompilerConfig instance:

    >>> from relationalai.config import create_config, CompilerConfig
    >>> cfg = create_config(
    ...     compiler=CompilerConfig(dry_run=True, strict=True, soft_type_errors=True),
    ... )
    """

    use_monotype_operators: bool = Field(default=False, description="Use monotype operators in compilation")
    show_corerel_errors: bool = Field(default=True, description="Show CoreRel error messages")
    dry_run: bool = Field(default=False, description="Run compilation in dry-run mode")
    use_value_types: bool = Field(default=False, description="Use value types in compilation")
    debug_hidden_keys: bool = Field(default=False, description="Debug hidden keys in compilation")
    wide_outputs: bool = Field(default=False, description="Use wide output format for query results")
    strict: bool = Field(default=False, description="Enable strict validation mode")
    soft_type_errors: bool = Field(default=False, description="Treat type errors as warnings instead of failures")
    safe_recursion: bool = Field(default=True, description="Enable safe recursion mode")

    # Experimental compiler optimizations
    use_inlined_intermediates: bool = Field(default=False, description="Use inlined intermediate results")
    inline_value_maps: bool = Field(default=False, description="Inline value maps in weaver")
    inline_entity_maps: bool = Field(default=False, description="Inline entity maps in weaver")


@include_in_docs
class ModelConfig(BaseModel):
    """Configure model execution defaults.

    This controls a few cross-cutting behaviors for running a
    :class:`~relationalai.semantics.frontend.base.Model`, including whether
    missing properties may be created implicitly and (for SQL execution) the
    schema used to install model-generated views and tables.

    Attributes
    ----------
    keep : bool
        Keep the model after execution.
    isolated : bool
        Run the model in isolated mode.
    nowait_durable : bool
        Do not wait for durable operations.
    schema_ : str, optional
        Install schema for model-generated SQL views and tables (config key is
        ``schema``).
    implicit_properties : bool
        Allow undeclared properties to be created on first access (when True).

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    model:
        schema: analytics
        implicit_properties: false
    ```

    Configure model settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(
    ...     model={"schema": "analytics", "implicit_properties": False},
    ... )

    Configure model settings using an explicit ModelConfig instance:

    >>> from relationalai.config import create_config, ModelConfig
    >>> cfg = create_config(
    ...     model=ModelConfig(schema_="analytics", implicit_properties=False),
    ... )
    """

    keep: bool = Field(default=True, description="Keep model after execution")
    isolated: bool = Field(default=True, description="Run model in isolated mode")
    nowait_durable: bool = Field(default=True, description="Don't wait for durable operations")
    reuse_model: bool = Field(default=True, description="Keep database after releasing index for potential reuse")
    schema_: str | None = Field(default=None, alias="schema", description="Schema for model-generated SQL views and tables")
    implicit_properties: bool = Field(default=True, description="Enable implicit properties for concepts")


@include_in_docs
class DebugConfig(BaseModel):
    """Configure debug server and error-reporting behavior.

    This controls local debugging features such as the debug server host/port
    and whether extra debug logs or full stack traces are shown.

    Attributes
    ----------
    enabled : bool
        Enable debug mode.
    host : str, optional
        Debug server host. If omitted, a default host is used.
    port : int
        Debug server port.
    show_debug_logs : bool
        Show debug log messages.
    show_full_traces : bool
        Show full stack traces in errors.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    debug:
        show_full_traces: true
    ```

    Configure debug settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(debug={"show_full_traces": True})
    """

    enabled: bool = Field(default=True, description="Enable debug mode")
    host: str | None = Field(default=None, description="Debug server host")
    port: int = Field(default=8080, description="Debug server port")
    show_debug_logs: bool = Field(default=False, description="Show debug log messages")
    show_full_traces: bool = Field(default=False, description="Show full stack traces in errors")


@include_in_docs
class JobsConfig(BaseModel):
    """Configure job execution behavior.

    Controls progress reporting and safety guardrails during job execution.

    Attributes
    ----------
    direct_access : bool, optional
        Override the top-level ``direct_access`` default for jobs only.
    print_progress : bool
        Print job progress information.
    print_progress_internal : bool
        Print internal job progress information (detailed).
    enable_guard_rails : bool
        Enable safety guardrails during job execution.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    jobs:
        direct_access: true
        print_progress: true
        enable_guard_rails: true
    ```

    Configure jobs settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(
    ...     connections={"sf": {...}},
    ...     jobs={
    ...         "direct_access": True,
    ...         "print_progress": True,
    ...         "enable_guard_rails": True,
    ...     },
    ... )

    Configure jobs settings using an explicit JobsConfig instance:

    >>> from relationalai.config import create_config, JobsConfig
    >>> cfg = create_config(
    ...     connections={"sf": {...}},
    ...     jobs=JobsConfig(
    ...         direct_access=True,
    ...         print_progress=True,
    ...         enable_guard_rails=True,
    ...     ),
    ... )
    """

    direct_access: bool | None = Field(
        default=None,
        description="Preferred Direct Access toggle for jobs. When set, overrides the top-level direct_access default.",
    )

    print_progress: bool = Field(
        default=True,
        description="Print job progress information"
    )

    print_progress_internal: bool = Field(
        default=False,
        description="Print internal job progress information"
    )

    enable_guard_rails: bool = Field(
        default=False,
        description="Enable job execution guardrails"
    )



"""Client middleware composition.

The SDK uses a shared runtime middleware pipeline for both SQL and HTTP execution.
This module translates high-level `ExecutionConfig` knobs into concrete middleware
instances (logging, metrics, retries).

Returned values:
- `middlewares`: ordered list to apply around the base executor
- `metrics`: an in-memory metrics sink (or a no-op sink)
"""

from __future__ import annotations

import logging
from typing import Any, Callable

from relationalai.config.config_fields import ExecutionConfig
from relationalai.runtime.middleware import (
    InMemoryMetricsSink,
    LoggingMiddleware,
    MetricsMiddleware,
    NoOpMetricsSink,
    RetryMiddleware,
    RetryPolicy,
)


def compose_middlewares(
    cfg: ExecutionConfig,
    *,
    is_retryable: Callable[[Exception, Any], bool],
    logger_name: str,
) -> tuple[list[Any], Any]:
    """Compose the middleware chain from generic execution config knobs."""
    enable_logging = cfg.logging
    enable_metrics = cfg.metrics
    enable_retries = cfg.retries.enabled

    retry_policy = RetryPolicy(
        max_attempts=cfg.retries.max_attempts,
        base_delay_s=cfg.retries.base_delay_s,
        max_delay_s=cfg.retries.max_delay_s,
        jitter=cfg.retries.jitter,
    )

    metrics = InMemoryMetricsSink() if enable_metrics else NoOpMetricsSink()
    logger = logging.getLogger(logger_name)

    middlewares: list[Any] = []
    if enable_logging:
        middlewares.append(LoggingMiddleware(logger=logger, enabled=True))
    if enable_metrics:
        middlewares.append(MetricsMiddleware(metrics=metrics))

    if enable_retries:
        middlewares.append(
            RetryMiddleware(
                policy=retry_policy,
                is_retryable=is_retryable,
                logger=logger,
                metrics=metrics,
                enabled=True,
            )
        )

    return middlewares, metrics




"""
The ``agent`` module provides a client for the Snowflake Cortex Agents REST API.
"""
import dataclasses
from dataclasses import dataclass
from typing import List, Optional, Dict, Any, Tuple
import json
import copy

import httpx

from relationalai.util.docutils import include_in_docs
from relationalai.util.error import Diagnostic, RAIException, exc

__include_in_docs__ = True


@include_in_docs
class APIResponseError(RAIException):
    """Custom exception for API response errors."""
    def __init__(self, message):
        diag = Diagnostic("APIResponseError", str(message), severity="error")
        super().__init__(diag)
        self.message = message


@include_in_docs
class CortexAgentApi:
    """Client for interacting with the Snowflake Cortex Agents REST API.

    Provides methods for creating, updating, describing, listing, deleting,
    and running Cortex agents. Requires the ``snowflake.cortex_user``
    database role. All requests timeout after 15 minutes.

    Parameters
    ----------
    client : httpx.Client
        Configured ``httpx.Client`` with authentication headers.
    """
    def __init__(self, client: httpx.Client):
        self._c = client

    @dataclass
    class Profile:
        """Agent profile configuration for visual display properties.

        Attributes
        ----------
        avatar : str or None
            URL or identifier for the agent's avatar image.
        color : str or None
            Color code for agent's visual theme.
        display_name : str or None
            Human-readable name shown in the UI.
        """
        avatar: Optional[str] = None
        color: Optional[str] = None
        display_name: Optional[str] = None

    @dataclass
    class Instructions:
        """Agent instructions for guiding behavior and responses.

        Attributes
        ----------
        response : str or None
            Instructions for how the agent should format and deliver responses.
        orchestration : str or None
            Instructions for how the agent should orchestrate tool usage and
            workflow.
        """
        response: Optional[str] = None
        orchestration: Optional[str] = None

    @dataclass
    class BudgetConfig:
        """Budget constraints for agent execution.

        If more than one constraint is specified, whichever is first hit will
        end the request.

        Attributes
        ----------
        seconds : int or None
            Time budget in seconds for agent execution.
        tokens : int or None
            Token budget for model consumption.
        """
        seconds: Optional[int] = None
        tokens: Optional[int] = None

    @dataclass
    class OrchestrationConfig:
        """Configuration for agent orchestration behavior.

        Attributes
        ----------
        budget : CortexAgentApi.BudgetConfig or None
            Budget constraints for the agent.
        """
        budget: Optional['CortexAgentApi.BudgetConfig'] = None

    @dataclass
    class Models:
        """Model configuration for agent orchestration.

        Specifies which LLM model the agent should use for orchestrating tool
        usage and generating responses.

        Supported orchestration model values as of Jan 2026:

        - ``claude-haiku-4-5``     (Claude Haiku 4.5)
        - ``claude-sonnet-4-5``    (Claude Sonnet 4.5)
        - ``claude-4-sonnet``      (Claude Sonnet 4)
        - ``claude-3-7-sonnet``    (Claude Sonnet 3.7)
        - ``claude-3-5-sonnet``    (Claude Sonnet 3.5)
        - ``openai-gpt-5``         (GPT 5)
        - ``openai-gpt-4.1``       (GPT 4.1)

        Attributes
        ----------
        orchestration : str or None
            The model identifier for agent orchestration.
        """
        orchestration: Optional[str] = None

    @dataclass
    class ToolSpec:
        """Specification for a tool that the agent can use.

        Tools enable agents to perform actions like searching, analyzing data,
        or calling custom functions.

        Attributes
        ----------
        type : str
            Tool type (e.g., ``"function"``, ``"cortex_analyst"``,
            ``"cortex_search"``, ``"web_search"``).
        name : str
            Unique identifier for the tool.
        description : str or None
            Human-readable description of what the tool does.
        input_schema : dict or None
            JSON schema defining the tool's input parameters.
        """
        type: str
        name: str
        description: Optional[str] = None
        input_schema: Optional[Dict[str, Any]] = None

    @dataclass
    class Tool:
        """Container for a tool specification.

        Attributes
        ----------
        tool_spec : CortexAgentApi.ToolSpec
            The tool specification.
        """
        tool_spec: 'CortexAgentApi.ToolSpec'

    @dataclass
    class ExecutionEnvironment:
        """Configuration for server-side tool execution environment.

        Defines the Snowflake warehouse and timeout settings for executing
        tools that require compute resources.

        Attributes
        ----------
        query_timeout : int or None
            Maximum execution time in seconds for queries.
        type : str or None
            Environment type identifier.
        warehouse : str or None
            Name of the Snowflake warehouse to use for execution.
        """
        query_timeout: Optional[int] = None
        type: Optional[str] = None
        warehouse: Optional[str] = None

    @dataclass
    class ToolResourceSpec:
        """Resource configuration for a specific tool.

        Defines the data sources, execution environment, and other resources
        that a tool needs to operate.

        Attributes
        ----------
        execution_environment : CortexAgentApi.ExecutionEnvironment or None
            Warehouse configuration for server-side execution.
        semantic_model_file : str or None
            Path to semantic model file for Cortex Analyst.
        semantic_view : str or None
            Fully qualified name of semantic view.
        identifier : str or None
            Resource identifier.
        name : str or None
            Resource name.
        type : str or None
            Resource type.
        """
        execution_environment: Optional['CortexAgentApi.ExecutionEnvironment'] = None
        semantic_model_file: Optional[str] = None
        semantic_view: Optional[str] = None
        identifier: Optional[str] = None
        name: Optional[str] = None
        type: Optional[str] = None

    @dataclass
    class AgentSpec:
        """Complete specification for a Cortex agent.

        Defines all configuration aspects of an agent including its identity,
        behavior, capabilities, and resource access.

        Attributes
        ----------
        name : str or None
            Agent identifier (must be unique within schema).
        profile : CortexAgentApi.Profile or None
            Visual and display configuration.
        models : CortexAgentApi.Models or None
            Model selection for orchestration.
        instructions : CortexAgentApi.Instructions or None
            Behavioral guidance for the agent.
        tools : list of CortexAgentApi.Tool or None
            Tools available to the agent.
        tool_resources : dict or None
            Resource configurations keyed by tool name.
        orchestration : CortexAgentApi.OrchestrationConfig or None
            Orchestration configuration including budget constraints.
        """
        name: Optional[str] = None
        profile: Optional['CortexAgentApi.Profile'] = None
        models: Optional['CortexAgentApi.Models'] = None
        instructions: Optional['CortexAgentApi.Instructions'] = None
        tools: Optional[List['CortexAgentApi.Tool']] = None
        tool_resources: Optional[Dict[str, 'CortexAgentApi.ToolResourceSpec']] = None
        orchestration: Optional['CortexAgentApi.OrchestrationConfig'] = None

        def set_semantic_view(self, tool_name: str, view_fqn: str) -> 'CortexAgentApi.AgentSpec':
            """Set the semantic view for a specific tool.

            Parameters
            ----------
            tool_name : str
                Name of the tool to configure.
            view_fqn : str
                Fully qualified name of the semantic view.

            Returns
            -------
            CortexAgentApi.AgentSpec
                Self for method chaining.
            """
            assert self.tool_resources is not None, "tool_resources must be initialized"
            assert tool_name in self.tool_resources, f"tool_name '{tool_name}' not in tool_resources"
            self.tool_resources[tool_name].semantic_view = view_fqn
            return self

        def copy(self) -> 'CortexAgentApi.AgentSpec':
            """Create a deep copy of this agent specification.

            Returns
            -------
            CortexAgentApi.AgentSpec
                Independent copy of this AgentSpec.
            """
            return copy.deepcopy(self)

        def merge(self, other: 'CortexAgentApi.AgentSpec') -> 'CortexAgentApi.AgentSpec':
            """Merge another AgentSpec into this one, creating a new combined specification.

            Non-None fields from the other spec override or extend this spec's
            fields. Tools and tool_resources are combined rather than replaced.

            Parameters
            ----------
            other : CortexAgentApi.AgentSpec
                AgentSpec to merge into this one.

            Returns
            -------
            CortexAgentApi.AgentSpec
                New AgentSpec with merged configuration.
            """
            result = self.copy()
            if other.name is not None:
                result.name = other.name
            if other.profile is not None:
                result.profile = other.profile
            if other.models is not None:
                result.models = other.models
            if other.instructions is not None:
                result.instructions = other.instructions
            if other.orchestration is not None:
                result.orchestration = other.orchestration
            if other.tools is not None:
                if result.tools is None:
                    result.tools = copy.deepcopy(other.tools)
                else:
                    result.tools = result.tools + copy.deepcopy(other.tools)
            if other.tool_resources is not None:
                if result.tool_resources is None:
                    result.tool_resources = copy.deepcopy(other.tool_resources)
                else:
                    result.tool_resources = {**result.tool_resources, **copy.deepcopy(other.tool_resources)}
            return result

        def to_dict(self) -> dict:
            """Convert this AgentSpec to a dictionary.

            Returns
            -------
            dict
                Dictionary representation suitable for JSON serialization.
            """
            return dataclasses.asdict(self)

        @staticmethod
        def from_dict(spec_data: dict) -> Optional['CortexAgentApi.AgentSpec']:
            """Parse an AgentSpec from a dictionary.

            Parameters
            ----------
            spec_data : dict
                Dictionary containing agent specification data.

            Returns
            -------
            CortexAgentApi.AgentSpec or None
                Parsed AgentSpec instance, or ``None`` if parsing fails.
            """
            # Parse name
            name = spec_data.get("name")

            # Parse profile
            profile = None
            if spec_data.get("profile"):
                profile = CortexAgentApi.Profile(
                    avatar=spec_data["profile"].get("avatar"),
                    color=spec_data["profile"].get("color"),
                    display_name=spec_data["profile"].get("display_name")
                )

            # Parse models
            models = None
            if spec_data.get("models"):
                models = CortexAgentApi.Models(
                    orchestration=spec_data["models"].get("orchestration")
                )

            # Parse instructions
            instructions = None
            if spec_data.get("instructions"):
                instructions = CortexAgentApi.Instructions(
                    response=spec_data["instructions"].get("response"),
                    orchestration=spec_data["instructions"].get("orchestration")
                )

            # Parse tools
            tools = None
            if spec_data.get("tools"):
                tools = []
                for tool_data in spec_data["tools"]:
                    tool_spec_data = tool_data.get("tool_spec", {})
                    tool_spec = CortexAgentApi.ToolSpec(
                        type=tool_spec_data["type"],
                        name=tool_spec_data["name"],
                        description=tool_spec_data.get("description"),
                        input_schema=tool_spec_data.get("input_schema")
                    )
                    tool = CortexAgentApi.Tool(tool_spec=tool_spec)
                    tools.append(tool)

            # Parse tool resources
            tool_resources = None
            if spec_data.get("tool_resources"):
                tool_resources = {}
                for resource_name, resource_data in spec_data["tool_resources"].items():
                    execution_env = None
                    if resource_data.get("execution_environment"):
                        env_data = resource_data["execution_environment"]
                        execution_env = CortexAgentApi.ExecutionEnvironment(
                            query_timeout=env_data.get("query_timeout"),
                            type=env_data.get("type"),
                            warehouse=env_data.get("warehouse")
                        )

                    resource_spec = CortexAgentApi.ToolResourceSpec(
                        execution_environment=execution_env,
                        semantic_model_file=resource_data.get("semantic_model_file"),
                        semantic_view=resource_data.get("semantic_view"),
                        identifier=resource_data.get("identifier"),
                        name=resource_data.get("name"),
                        type=resource_data.get("type")
                    )
                    tool_resources[resource_name] = resource_spec

            # Parse orchestration
            orchestration = None
            if spec_data.get("orchestration"):
                budget = None
                if spec_data["orchestration"].get("budget"):
                    budget_data = spec_data["orchestration"]["budget"]
                    budget = CortexAgentApi.BudgetConfig(
                        seconds=budget_data.get("seconds"),
                        tokens=budget_data.get("tokens")
                    )
                orchestration = CortexAgentApi.OrchestrationConfig(budget=budget)

            return CortexAgentApi.AgentSpec(
                name=name,
                profile=profile,
                models=models,
                instructions=instructions,
                tools=tools,
                tool_resources=tool_resources,
                orchestration=orchestration
            )

    @include_in_docs
    def create(self, database: str, schema: str, request: AgentSpec, create_mode: str = "orReplace") -> None:
        """Create an agent.

        Parameters
        ----------
        database : str
            Database name.
        schema : str
            Schema name.
        request : AgentSpec
            Agent specification.
        create_mode : str
            One of ``"orReplace"``, ``"errorIfExists"``, or ``"ifNotExists"``.
            Default: ``"orReplace"``.

        Returns
        -------
        None
        """
        valid_modes = ["orReplace", "errorIfExists", "ifNotExists"]
        if create_mode not in valid_modes:
            exc(
                "InvalidCreateMode",
                f"create_mode must be one of {valid_modes}, got: {create_mode}",
            )

        response = self._c.post(
            url=f"/api/v2/databases/{database}/schemas/{schema}/agents?createMode={create_mode}",
            json=dataclasses.asdict(request),
        )
        # {"status": "..."}
        response.raise_for_status()

    @include_in_docs
    def update(self, database: str, schema: str, name: str, request: AgentSpec) -> None:
        """Update an existing agent's configuration.

        Modifies the properties of an existing agent. Fields not included in
        the request are left unchanged.

        Parameters
        ----------
        database : str
            Database name.
        schema : str
            Schema name.
        name : str
            Agent name to update.
        request : AgentSpec
            Updated agent specification.

        Returns
        -------
        None

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails.
        """
        response = self._c.put(
            url=f"/api/v2/databases/{database}/schemas/{schema}/agents/{name}",
            json=dataclasses.asdict(request),
        )
        response.raise_for_status()

    @dataclass
    class DescribeAgentResponse:
        """Response containing comprehensive agent details.

        Attributes
        ----------
        agent_spec_dict : dict
            Raw dictionary of agent specification.
        created_on : str
            Timestamp of agent creation.
        database_name : str
            Database containing the agent.
        name : str
            Agent name.
        owner : str
            User or role that owns the agent.
        schema_name : str
            Schema containing the agent.
        agent_spec : CortexAgentApi.AgentSpec or None
            Parsed agent specification.
        """
        agent_spec_dict: Dict[str, Any]
        created_on: str
        database_name: str
        name: str
        owner: str
        schema_name: str
        agent_spec: Optional['CortexAgentApi.AgentSpec'] = None

    @include_in_docs
    def describe(self, database: str, schema: str, name: str) -> DescribeAgentResponse:
        """Retrieve comprehensive details about a specific agent.

        Returns complete agent metadata including configuration, ownership,
        and creation information.

        Parameters
        ----------
        database : str
            Database name.
        schema : str
            Schema name.
        name : str
            Agent name.

        Returns
        -------
        DescribeAgentResponse
            Full agent details.

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails or agent doesn't exist.
        """
        response = self._c.get(
            url=f"/api/v2/databases/{database}/schemas/{schema}/agents/{name}",
        )
        response.raise_for_status()
        data = response.json()

        agent_spec_dict = json.loads(data['agent_spec'])
        parsed_spec = CortexAgentApi.AgentSpec.from_dict(agent_spec_dict)
        if parsed_spec is not None and not parsed_spec.name:
            parsed_spec.name = name

        return CortexAgentApi.DescribeAgentResponse(
            agent_spec_dict=agent_spec_dict,
            created_on=data['created_on'],
            database_name=data['database_name'],
            name=data['name'],
            owner=data['owner'],
            schema_name=data['schema_name'],
            agent_spec=parsed_spec
        )

    @dataclass
    class ListAgentsResponseInner:
        """Summary information for a single agent in a list response.

        Attributes
        ----------
        name : str
            Agent name.
        comment : str or None
            Optional comment/description.
        created_on : str
            Timestamp of agent creation.
        database_name : str
            Database containing the agent.
        owner : str
            User or role that owns the agent.
        profile : CortexAgentApi.Profile or None
            Agent's visual configuration.
        schema_name : str
            Schema containing the agent.
        """
        name: str
        comment: Optional[str]
        created_on: str
        database_name: str
        owner: str
        profile: Optional['CortexAgentApi.Profile']
        schema_name: str

    @dataclass
    class ListAgentsResponse:
        """Response containing a list of agent summaries.

        Attributes
        ----------
        agents : list of CortexAgentApi.ListAgentsResponseInner
            Agent summary objects.
        """
        agents: List['CortexAgentApi.ListAgentsResponseInner']

    @include_in_docs
    def list(self, database: str, schema: str) -> ListAgentsResponse:
        """List all agents in a schema.

        Returns summary information for all agents in the specified database
        and schema.

        Parameters
        ----------
        database : str
            Database name.
        schema : str
            Schema name.

        Returns
        -------
        ListAgentsResponse
            Agent summaries.

        Raises
        ------
        httpx.HTTPStatusError
            If the request fails.
        """
        response = self._c.get(
            url=f"/api/v2/databases/{database}/schemas/{schema}/agents",
        )
        response.raise_for_status()
        data = response.json()
        agents = []
        for agent_data in data:
            profile = None
            if agent_data.get('profile'):
                profile = CortexAgentApi.Profile(**agent_data['profile'])

            agent = CortexAgentApi.ListAgentsResponseInner(
                name=agent_data['name'],
                comment=agent_data.get('comment'),
                created_on=agent_data['created_on'],
                database_name=agent_data['database_name'],
                owner=agent_data['owner'],
                profile=profile,
                schema_name=agent_data['schema_name']
            )
            agents.append(agent)
        return CortexAgentApi.ListAgentsResponse(agents=agents)

    @include_in_docs
    def delete(self, database: str, schema: str, name: str, silent=True):
        """Delete an agent.

        Permanently removes an agent from the database. The behavior when the
        agent doesn't exist depends on the ``silent`` parameter.

        Parameters
        ----------
        database : str
            Database name.
        schema : str
            Schema name.
        name : str
            Agent name to delete.
        silent : bool
            If ``True``, suppress errors when agent doesn't exist.
            Default: ``True``.

        Returns
        -------
        httpx.Response
            The HTTP response object.

        Raises
        ------
        httpx.HTTPStatusError
            If ``silent=False`` and the request fails.
        """
        response = self._c.delete(
            url=f"/api/v2/databases/{database}/schemas/{schema}/agents/{name}",
        )
        if not silent:
            response.raise_for_status()
        return response

    @include_in_docs
    def run(self, database: str, schema: str, name: str, thread_id: int, parent_message_id: int, prompt: str) -> Tuple[dict, int]:
        """Execute an agent with a user prompt in a conversation thread.

        Sends a message to the agent and streams back the response. Processes
        server-sent events to extract the agent's response and updated message
        ID.

        Parameters
        ----------
        database : str
            Database name.
        schema : str
            Schema name.
        name : str
            Agent name to run.
        thread_id : int
            Conversation thread identifier.
        parent_message_id : int
            ID of the previous message in the thread.
        prompt : str
            User's message/prompt for the agent.

        Returns
        -------
        tuple of (dict, int)
            ``(response_dict, new_parent_message_id)`` where
            ``response_dict`` contains the agent's response with ``"content"``
            and ``"role"`` fields, and ``new_parent_message_id`` is the message
            ID for the next interaction.

        Raises
        ------
        APIResponseError
            If the agent returns an error.
        httpx.HTTPStatusError
            If the HTTP request fails.
        AssertionError
            If the response doesn't contain expected content.
        """
        error_dict = None
        with self._c.stream(
                "POST",
                url=f"/api/v2/databases/{database}/schemas/{schema}/agents/{name}:run",
                json={
                    "thread_id": thread_id,
                    "parent_message_id": parent_message_id,
                    "messages": [{
                        "role": "user",
                        "content": [{
                            "type": "text",
                            "text": prompt
                        }]
                    }],
                },
        ) as response:
            if not response.is_success:
                response.raise_for_status()
            data = None
            event = None
            for line in response.iter_lines():
                # print(line)
                if line.startswith("event: "):
                    event = line
                elif line.startswith("data: "):
                    if event == "event: error":
                        error_dict = json.loads(line[6:])
                    if event == "event: response":
                        data = json.loads(line[6:])
                        break
                    if event == "event: metadata":
                        # data: {"metadata": {"message_id": 37856360561, "role": "assistant"}}
                        data = json.loads(line[6:])
                        if data["metadata"]["role"] == "assistant":
                            parent_message_id = data["metadata"]["message_id"]
        if error_dict:
            raise APIResponseError(error_dict)
        assert data is not None, "Response data should not be None"
        assert "content" in data and data["content"] is not None and len(data["content"]) > 0, "Response must have content"
        data["role"] = "assistant"
        return data, parent_message_id



"""Reasoner configuration sections for :func:`relationalai.config.Config`.

Most projects configure reasoner defaults in a YAML file (typically
``raiconfig.yaml``) under the ``reasoners`` section (for example
``reasoners.logic.size`` and ``reasoners.logic.auto_suspend_mins``). This
module defines the Pydantic models that validate those settings, along with the
allowed reasoner size strings for each deployment platform.

You can also configure everything in code by passing dictionaries to
the :func:`~relationalai.config.create_config` factory function (useful for programmatic
setup or for overrides in tests).

Start with :class:`~relationalai.config.config_reasoners_fields.ReasonersConfig`
(the top-level ``reasoners`` section) for configuring base reasoner API settings.
Shared fields live in :class:`~relationalai.config.config_reasoners_fields.ReasonerConfig`,
and logic reasoner-specific options live in
:class:`~relationalai.config.config_reasoners_fields.LogicReasonerConfig`.
"""

from __future__ import annotations

from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field

from ..util.docutils import include_in_docs


# include this module in documentation
__include_in_docs__ = True


# Engine size type aliases
# Internal development sizes
InternalEngineSize = Literal["XS", "S", "M", "L"]
"""Internal-only reasoner size values.

These sizes should not be used by end users and are typically hidden unless
:attr:`~relationalai.config.config_reasoners_fields.ReasonersConfig.show_all_sizes` is enabled.
"""

# AWS Snowflake sizes
AWSEngineSize = Literal["HIGHMEM_X64_S", "HIGHMEM_X64_M", "HIGHMEM_X64_L", "GPU_NV_S"]
"""Valid reasoner size values on AWS-backed deployments.

Use one of these strings for :attr:`~relationalai.config.config_reasoners_fields.ReasonerConfig.size`.
"""

# Azure sizes
AzureEngineSize = Literal["HIGHMEM_X64_S", "HIGHMEM_X64_M", "HIGHMEM_X64_SL"]
"""Valid reasoner size values on Azure-backed deployments.

Use one of these strings for :attr:`~relationalai.config.config_reasoners_fields.ReasonerConfig.size`.
"""

# All valid engine sizes across platforms
EngineSize = InternalEngineSize | AWSEngineSize | AzureEngineSize
"""All supported reasoner size values.

This is the type accepted by :attr:`~relationalai.config.config_reasoners_fields.ReasonerConfig.size`.
Internal/dev sizes are included but are typically hidden unless
:attr:`~relationalai.config.config_reasoners_fields.ReasonersConfig.show_all_sizes` is enabled.
"""

# LQP Semantics Version
# When the logic engine introduces breaking changes, this version is bumped
# [2026-01-09] Version "1" opts into hard validation errors from the engine
LqpSemanticsVersion = Literal["1"]


@include_in_docs
class ReasonerLogicLqpConfig(BaseModel):
    """Configure LQP-specific behavior for the logic reasoner.

    Most users should leave this unset. Only set it if a RelationalAI release
    note or support instruction asks you to opt into a specific semantics
    version.

    Attributes
    ----------
    semantics_version : {"1"}, optional
        Semantics version to opt into.
    """

    semantics_version: LqpSemanticsVersion | None = Field(
        default=None,
        description="LQP semantics version for breaking changes (e.g., '1' for hard validation errors)"
    )


@include_in_docs
class ReasonerConfig(BaseModel):
    """Common configuration fields for a reasoner.

    This model defines the fields shared by all reasoner types. You set these
    fields under one of the reasoner subsections in ``raiconfig.yaml``:

    - ``reasoners.logic`` (logic reasoner, see
      :class:`~relationalai.config.config_reasoners_fields.LogicReasonerConfig`
      for additional logic-specific settings)
    - ``reasoners.predictive`` (predictive reasoner)
    - ``reasoners.prescriptive`` (prescriptive reasoner)

    Attributes
    ----------
    name : str, optional
        Optional name for the reasoner instance.
    size : str
        Compute size (for example, ``"HIGHMEM_X64_S"``).
        See :attr:`~relationalai.config.config_reasoners_fields.EngineSize` for valid size values.
    query_timeout_mins : int, optional
        Abort queries that exceed this duration.
    auto_suspend_mins : int
        Auto-suspend the reasoner after N minutes of inactivity.
    await_storage_vacuum : bool, optional
        Whether to wait for storage vacuum work to finish before suspending. Defaults to ``None``.
    settings : dict, optional
        Backend-specific settings payload.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``) configuring only the logic reasoner:

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    reasoners:
        logic:
            size: HIGHMEM_X64_S
            query_timeout_mins: 30
    ```

    Any omitted subsections (for example ``reasoners.predictive``) keep their
    built-in defaults.

    Configure reasoner defaults using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(reasoners={
    ...     "logic": {
    ...         "size": "HIGHMEM_X64_S",
    ...         "query_timeout_mins": 30,
    ...     }
    ... })

    Configure reasoner defaults using an explicit ReasonerConfig instance:

    >>> from relationalai.config import create_config, ReasonerConfig
    >>> cfg = create_config(
    ...     reasoners={"predictive": ReasonerConfig(size="HIGHMEM_X64_S", query_timeout_mins=30)},
    ... )
    """

    model_config = ConfigDict(extra="allow")

    name: str | None = Field(
        default=None,
        description="Optional name for the reasoner instance"
    )

    size: EngineSize = Field(
        default="HIGHMEM_X64_S",
        description="Compute size (e.g., 'HIGHMEM_X64_S', 'HIGHMEM_X64_M')",
        examples=["HIGHMEM_X64_S", "HIGHMEM_X64_M", "HIGHMEM_X64_L"]
    )

    query_timeout_mins: int | None = Field(
        default=None,
        description="Query timeout in minutes - aborts queries that exceed this duration"
    )

    auto_suspend_mins: int = Field(
        default=60,
        description="Auto-suspend reasoner after N minutes of inactivity"
    )

    await_storage_vacuum: bool | None = Field(
        default=None,
        description="Wait for storage vacuum work to finish before suspending the reasoner",
    )

    settings: dict[str, Any] | None = Field(
        default=None,
        description="Optional reasoner settings payload (backend-dependent)",
    )


@include_in_docs
class LogicReasonerConfig(ReasonerConfig):
    """Configure rule-execution defaults for the logic reasoner.

    Set these fields under ``reasoners.logic`` to control how the logic reasoner
    runs rules (for example, whether to use LQP) and to opt into an LQP
    semantics version when required.

    Attributes
    ----------
    use_lqp : bool
        Whether to use LQP for rule execution.
    incremental_maintenance : str, optional
        Incremental maintenance strategy string.
        Can be set to ``"on"``, ``"auto"``, or ``"off"`` (default).
    emit_constraints : bool
        Whether to emit constraints during execution.
    lqp : ReasonerLogicLqpConfig
        LQP-specific options, such as ``semantics_version``.
        See :class:`~relationalai.config.config_reasoners_fields.ReasonerLogicLqpConfig` for details.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    reasoners:
        logic:
            incremental_maintenance: "on"
    ```

    Configure logic reasoner settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(reasoners={"logic": {"use_lqp": True, "lqp": {"semantics_version": "1"}}})

    Configure logic reasoner settings using an explicit LogicReasonerConfig instance:

    >>> from relationalai.config import create_config, LogicReasonerConfig
    >>> cfg = create_config(
    ...     reasoners={"logic": LogicReasonerConfig(use_lqp=True, lqp={"semantics_version": "1"})},
    ... )
    """

    use_lqp: bool = Field(
        default=True,
        description="Use LQP for reasoner rule execution"
    )

    incremental_maintenance: str | None = Field(
        default="off",
        description="Incremental maintenance strategy"
    )

    emit_constraints: bool = Field(
        default=False,
        description="Emit constraints during execution"
    )

    lqp: ReasonerLogicLqpConfig = Field(
        default_factory=ReasonerLogicLqpConfig,
        description="LQP-specific configuration"
    )


@include_in_docs
class PrescriptiveReasonerConfig(ReasonerConfig):
    """Defaults for the prescriptive reasoner."""

    auto_suspend_mins: int = Field(
        default=30,
        description="Auto-suspend reasoner after N minutes of inactivity",
    )


@include_in_docs
class ReasonersConfig(BaseModel):
    """Configure defaults for the reasoners API.

    This section lives under ``reasoners`` in ``raiconfig.yaml`` and controls the
    reasoners backend, polling behavior for long-running operations, and default
    settings for each reasoner type (``logic``, ``predictive``, ``prescriptive``).

    Attributes
    ----------
    backend : {"sql", "direct_access"}
        Backend used for reasoners APIs.
    show_all_sizes : bool
        Show all available reasoner sizes (includes internal/dev sizes).
    poll_initial_delay_s : float
        Initial delay between polls when waiting for operations to complete.
    poll_overhead_rate : float
        Exponential backoff growth rate per poll (for example, ``0.2`` means +20% per poll).
    poll_max_delay_s : float
        Maximum poll delay.
    logic : LogicReasonerConfig
        Defaults for the logic reasoner.
    predictive : ReasonerConfig
        Defaults for the predictive reasoner.
    prescriptive : ReasonerConfig
        Defaults for the prescriptive reasoner.

    Examples
    --------
    Minimal YAML (in ``raiconfig.yaml``):

    ```yaml
    default_connection: sf
    connections:
        sf:
            type: snowflake
            # ...
    reasoners:
        direct_access: true
        poll_max_delay_s: 2.0
        logic:
            size: HIGHMEM_X64_S
            query_timeout_mins: 30
    ```

    Configure reasoners settings using a dict:

    >>> from relationalai.config import create_config
    >>> cfg = create_config(reasoners={
    ...     "poll_max_delay_s": 2.0,
    ...     "logic": {"size": "HIGHMEM_X64_S", "query_timeout_mins": 30},
    ... })

    Configure reasoners settings using an explicit ReasonersConfig instance:

    >>> from relationalai.config import create_config, ReasonersConfig, LogicReasonerConfig
    >>> cfg = create_config(
    ...     reasoners=ReasonersConfig(
    ...         poll_max_delay_s=2.0,
    ...         logic=LogicReasonerConfig(size="HIGHMEM_X64_S", query_timeout_mins=30),
    ...     ),
    ... )
    """

    backend: Literal["sql", "direct_access"] | None = Field(
        default=None,
        description="Legacy compatibility field. Prefer `reasoners.direct_access` or the top-level `direct_access` setting.",
    )

    direct_access_base_url: str | None = Field(
        default=None,
        description="Legacy compatibility field for Direct Access base URL. Newer DA wiring resolves service endpoints dynamically.",
    )

    direct_access: bool | None = Field(
        default=None,
        description="Preferred Direct Access toggle for reasoners. When set, overrides the top-level direct_access default.",
    )

    # Legacy note: there is also `config.engine.show_all_sizes` (engine terminology).
    # The reasoners surface uses the same knob, but names it here to avoid leaking
    # backend "engine" concepts into the reasoners API.
    show_all_sizes: bool = Field(
        default=False,
        description="Show all available reasoner sizes (includes internal/dev sizes).",
    )

    # Polling defaults used by `wait_until_ready(...)`, `ReasonerOperation.wait(...)`,
    # and other "wait until done" helpers. These are *not* retry policy knobs; they
    # control how aggressively we poll backend state during long-running operations.
    poll_initial_delay_s: float = Field(default=0.05, description="Initial poll delay (seconds)")
    poll_overhead_rate: float = Field(default=0.2, description="Exponential backoff growth rate per poll (0.2 => +20%)")
    poll_max_delay_s: float = Field(default=2.0, description="Maximum poll delay (seconds)")

    logic: LogicReasonerConfig = Field(
        default_factory=LogicReasonerConfig,
        description="Logic reasoner configuration"
    )

    predictive: ReasonerConfig = Field(
        default_factory=ReasonerConfig,
        description="Predictive reasoner configuration"
    )

    prescriptive: PrescriptiveReasonerConfig = Field(
        default_factory=PrescriptiveReasonerConfig,
        description="Prescriptive reasoner configuration"
    )



"""
Integer manipulation functions.

.. deprecated:: 0.13.0
   This module is deprecated. Use `std.numbers` instead.

This module provides functions for working with integers including type coercion
and parsing from strings.
"""
from __future__ import annotations

from ..frontend.base import Expression, Variable
from ..frontend.core import Integer
from relationalai.util.docutils import include_in_docs

from . import StringValue, IntegerValue, numbers, _deprecated_library

__include_in_docs__ = True


@include_in_docs
def int64(value: IntegerValue) -> Variable:
    """
    Coerce a number to Int64.

    .. deprecated:: 0.13.0
       Use `std.numbers` functions instead.

    Parameters
    ----------
    value: IntegerValue
        The integer value to coerce.

    Returns
    -------
    Variable
        A Variable representing the Int64 value.
    """
    _warning()
    return Integer(value)

@include_in_docs
def int128(value: IntegerValue) -> Variable:
    """
    Coerce a number to Int128.

    .. deprecated:: 0.13.0
       Use `std.numbers` functions instead.

    Parameters
    ----------
    value: IntegerValue
        The integer value to coerce.

    Returns
    -------
    Variable
        A Variable representing the Int128 value.
    """
    _warning()
    return Integer(value)

@include_in_docs
def parse_int64(value: StringValue) -> Expression:
    """
    Parse a string into an Int64.

    .. deprecated:: 0.13.0
       Use `numbers.parse_number(value, 19, 0) <std.numbers.parse_number>` instead.

    Parameters
    ----------
    value: StringValue
        The string value to parse.

    Returns
    -------
    Expression
        An Expression computing the parsed Int64.
    """
    _warning(f"numbers.parse_number({value}, 19, 0)")
    return numbers.parse_number(value, 19, 0)

@include_in_docs
def parse_int128(value: StringValue) -> Expression:
    """
    Parse a string into an Int128.

    .. deprecated:: 0.13.0
       Use `numbers.parse_number(value, 38, 0) <std.numbers.parse_number>` instead.

    Parameters
    ----------
    value: StringValue
        The string value to parse.

    Returns
    -------
    Expression
        An Expression computing the parsed Int128.
    """
    _warning(f"numbers.parse_number({value}, 38, 0)")
    return numbers.parse_number(value, 38, 0)

@include_in_docs
def parse(value: StringValue) -> Expression:
    """
    Parse a string into an Int128 (alias for parse_int128).

    .. deprecated:: 0.13.0
       Use `numbers.parse_number(value, 38, 0) <std.numbers.parse_number>` instead.

    Parameters
    ----------
    value: StringValue
        The string value to parse.

    Returns
    -------
    Expression
        An Expression computing the parsed Int128.
    """
    return parse_int128(value)


#--------------------------------------------------
# Implementation
#--------------------------------------------------

def _warning(msg: str|None = None):
    _deprecated_library("std.integers", "std.numbers", msg)



from __future__ import annotations

import json
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Iterable, Mapping

import rich
from rich import box as rich_box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table

if TYPE_CHECKING:
    from relationalai.config.config import Config
    from relationalai.config.connections import ConnectionConfig

from relationalai.services.reasoners.models import ReasonerType


def format_state_with_color(state: str) -> str:
    """Format state with colors (mirrors upstream CLI `format_state_with_color`)."""
    if not state:
        return ""
    state_upper = state.upper()
    # Success / terminal-ok states (jobs + operations).
    if state_upper in ("COMPLETED", "SUCCEEDED"):
        return f"[green]{state_upper}[/green]"
    if state_upper == "READY":
        return f"[green]{state_upper}[/green]"
    if state_upper == "SUSPENDED":
        return f"[yellow]{state_upper}[/yellow]"
    if state_upper in ("CANCELED", "CANCELLED"):
        return f"[yellow]{state_upper}[/yellow]"
    if state_upper in ("PENDING", "SYNCING", "PROCESSING", "RUNNING", "CANCELLING"):
        return f"[bold yellow]{state_upper}[/bold yellow]"
    if state_upper in ("ABORTED", "QUARANTINED", "GONE", "FAILED", "ERROR", "DELETED"):
        return f"[red]{state_upper}[/red]"
    return state_upper


def format_value(value: Any) -> str:
    """Format values for table display (mirrors upstream CLI `format_value`)."""
    if value is None:
        return "N/A"
    if isinstance(value, (int, float, bool)):
        return f"{value}"
    if isinstance(value, list):
        return ", ".join(map(str, value))
    if isinstance(value, datetime):
        return value.strftime("%Y-%m-%d %H:%M:%S")
    if isinstance(value, timedelta):
        # Keep it simple here; reasoners tables typically don't show timedeltas.
        return str(value)
    return str(value)


def _format_duration_ms(duration_ms: int | float | None) -> str:
    """Format a duration in milliseconds into a compact human form."""
    if duration_ms is None:
        return "N/A"
    try:
        ms = float(duration_ms)
    except Exception:
        return str(duration_ms)
    if ms < 1000:
        return f"{int(ms)}ms" if ms.is_integer() else f"{ms}ms"
    s = ms / 1000.0
    # Keep one decimal only when needed (e.g. 1.5s), otherwise show integer seconds.
    if abs(s - round(s)) < 1e-9:
        return f"{int(round(s))}s"
    out = f"{s:.1f}".rstrip("0").rstrip(".")
    return f"{out}s"


_KNOWN_REASONER_TYPES = {ReasonerType.LOGIC, ReasonerType.PRESCRIPTIVE, ReasonerType.PREDICTIVE}


def _get(obj: Any, key: str, default: Any = "") -> Any:
    if isinstance(obj, Mapping):
        return obj.get(key, default)
    return getattr(obj, key, default)


def _format_reasoner_type(reasoner_type: str) -> str:
    t = (reasoner_type or "").upper()
    from relationalai.tools.cli.job_types import JOB_TYPE_LABELS

    if t in JOB_TYPE_LABELS:
        return JOB_TYPE_LABELS[t]
    if t in _KNOWN_REASONER_TYPES:
        return ReasonerType.get_label(t)
    return reasoner_type or ""


def show_reasoners(reasoners: Iterable[Any]) -> None:
    """Show reasoners in a table matching upstream reasoners table layout."""
    items = list(reasoners)
    if not len(items):
        return

    table = Table(show_header=True, border_style="dim", header_style="bold", box=rich_box.SIMPLE_HEAD)
    table.add_column("#")
    table.add_column("Name")
    table.add_column("Type")
    table.add_column("Size")
    table.add_column("State")
    table.add_column("Created By")
    table.add_column("Created On")

    for index, r in enumerate(items):
        r_type = _get(r, "type", "") or ""
        type_display = _format_reasoner_type(str(r_type))
        created_on = format_value(_get(r, "created_on", None))
        state_display = format_state_with_color(str(_get(r, "state", "") or ""))
        table.add_row(
            f"{index+1}",
            str(_get(r, "name", "")),
            type_display,
            str(_get(r, "size", "") or ""),
            state_display,
            str(_get(r, "created_by", "") or ""),
            created_on,
        )

    rich.print(table)


def show_reasoner(reasoner: Any) -> None:
    """Convenience wrapper for printing a single reasoner."""
    show_reasoners([reasoner])


def show_jobs(jobs: Iterable[Any], *, include_details: bool = False) -> None:
    """Show jobs in a stable, readable table.

    When `include_details=True`, shows extra columns populated from `job.details`
    (if present) such as database/read_only/timeout_ms.
    """
    items = list(jobs)
    if not len(items):
        return

    table = Table(show_header=True, border_style="dim", header_style="bold", box=rich_box.SIMPLE_HEAD)
    table.add_column("#")
    table.add_column("ID", no_wrap=True)
    table.add_column("Type")
    table.add_column("State")
    table.add_column("Reasoner")
    table.add_column("Created By")
    table.add_column("Created On")
    table.add_column("Finished At")
    table.add_column("Duration")
    table.add_column("Abort Reason")
    if include_details:
        table.add_column("Database")
        table.add_column("Read Only")
        table.add_column("Timeout (ms)")

    for index, j in enumerate(items):
        t_display = _format_reasoner_type(str(_get(j, "reasoner_type", "") or ""))
        state_display = format_state_with_color(str(_get(j, "state", "") or ""))
        details = _get(j, "details", None)
        details_d: dict[str, Any] = dict(details) if isinstance(details, Mapping) else {}
        dur_any = _get(j, "duration", None)
        duration_display = (
            _format_duration_ms(dur_any)
            if (dur_any is None or isinstance(dur_any, (int, float)))
            else format_value(dur_any)
        )
        table.add_row(
            f"{index+1}",
            str(_get(j, "id", "")),
            t_display,
            state_display,
            format_value(_get(j, "name", None)),
            format_value(_get(j, "created_by", None)),
            format_value(_get(j, "created_on", None)),
            format_value(_get(j, "finished_at", None)),
            duration_display,
            format_value(_get(j, "abort_reason", None)),
            *(
                [
                    format_value(details_d.get("database_name")),
                    format_value(details_d.get("read_only")),
                    format_value(details_d.get("timeout_ms")),
                ]
                if include_details
                else []
            ),
        )

    rich.print(table)


def show_job(job: Any) -> None:
    """Convenience wrapper for printing a single job."""
    show_jobs([job])


def show_connect_panel(config: Config, console: Console) -> ConnectionConfig:
    """Print the rai connect info panel and return the resolved default connection."""
    from relationalai.tools.cli.error_handlers import format_connection_info

    connection = config.get_default_connection()

    info_table = Table(show_header=False, box=None, padding=(0, 2, 0, 0), expand=True)
    info_table.add_column(style="dim cyan", no_wrap=True)
    info_table.add_column(style="white")

    source_field = type(config).model_fields.get("source")
    config_source = source_field.default if source_field is not None else None
    if config_source:
        info_table.add_row("Config", config_source)
    if config.active_profile:
        info_table.add_row("Profile", config.active_profile)

    if config_source or config.active_profile:
        info_table.add_row("", "")  # spacer

    for label, value in format_connection_info(connection):
        info_table.add_row(label, value)

    console.print(Panel(info_table, title="[bold cyan]rai connect[/bold cyan]", border_style="cyan", padding=(1, 2)))
    return connection


def print_job_expansions(job: Any) -> None:
    """Print expanded job sections when present (mirrors scripts/jobs.py UX)."""
    if getattr(job, "details", None) is not None:
        rich.print("\n[bold]DETAILS[/bold]")
        rich.print(json.dumps(getattr(job, "details"), indent=2, default=str))
    if getattr(job, "artifacts", None) is not None:
        rich.print("\n[bold]ARTIFACTS[/bold]")
        rich.print(json.dumps(getattr(job, "artifacts"), indent=2, default=str))
    if getattr(job, "problems", None) is not None:
        rich.print("\n[bold]PROBLEMS[/bold]")
        rich.print(json.dumps(getattr(job, "problems"), indent=2, default=str))
    if getattr(job, "events", None) is not None:
        ev = getattr(job, "events")
        rich.print("\n[bold]EVENTS[/bold]")
        rich.print(
            json.dumps(
                {"events": getattr(ev, "events", []), "continuation_token": getattr(ev, "continuation_token", None)},
                indent=2,
                default=str,
            )
        )
    if getattr(job, "raw", None) is not None:
        rich.print("\n[bold]RAW[/bold]")
        rich.print(json.dumps(getattr(job, "raw"), indent=2, default=str))




"""
The ``queries`` module provides abstractions for managing pre-defined RAI queries
that Cortex agents can discover and execute.
"""
import inspect
from abc import ABC
from typing import Callable, Dict, Any, Optional, Union
import warnings

import pandas
import relationalai.semantics as rai

from relationalai.util.docutils import include_in_docs

__include_in_docs__ = True


@include_in_docs
class Queries(ABC):
    """Abstract base class for managing pre-defined RAI queries.

    Queries provide a registry of pre-built query functions that Cortex agents
    can discover and execute. This allows domain experts to define common
    analytical queries that agents can use without writing RAI code.

    Implement this interface to create custom query providers, or use
    ``QueryCatalog`` for standard Python function-based queries.
    """

    @include_in_docs
    def get_available_queries(self) -> dict:
        """Return metadata about available queries.

        Returns
        -------
        dict
            Dictionary with the following keys:

            queries : dict
                Mapping from a query ID to a query metadata dictionary.
                Each query metadata dictionary contains:

                id : str
                    Unique query identifier.
                description : str
                    Human-readable description of the query.
                args : dict
                    Mapping from parameter name to parameter type.

            query_instructions : str
                Instructions describing how to use the catalog.
        """
        raise NotImplementedError()

    @include_in_docs
    def explain(self, query_id: str) -> Optional[str]:
        """Return source code for a query by ID, or None if not found.

        Parameters
        ----------
        query_id : str
            Unique identifier for the query to explain.

        Returns
        -------
        str or None
            Source code string, or ``None`` if the query ID is not found.
        """
        raise NotImplementedError()

    @include_in_docs
    def execute(self, query_id: str, args: dict) -> Union[rai.Fragment, pandas.DataFrame]:
        """Execute a specific query by ID with provided arguments.

        Parameters
        ----------
        query_id : str
            Unique identifier for the query to execute (e.g.,
            ``"top_customers"``).
        args : dict
            Dictionary mapping parameter names to values (e.g.,
            ``{"limit": 5, "days": 30}``).

        Returns
        -------
        Union[rai.Fragment, pandas.DataFrame]
            Query results.

        Raises
        ------
        KeyError
            If the ``query_id`` is not found in registered queries.
        """
        raise NotImplementedError()


@include_in_docs
class QueryCatalog(Queries):
    """Query provider that registers Python functions as executable queries.

    Extracts metadata from Python callables (functions, methods) including
    their docstrings, type annotations, and source code. This metadata is
    exposed to Cortex agents for query discovery, while the functions remain
    executable for query execution.

    Query functions should follow these conventions:

    1. Return type: Must return either ``rai.Fragment`` or ``pandas.DataFrame``
    2. Parameters: Can have any parameters with type annotations (recommended)
    3. Docstring: Should include a clear description of what the query does
    4. Function name: Will be used as the query ID (e.g., ``"top_customers"``)

    Parameters
    ----------
    *query_funcs : Callable
        Variable number of Python callables to register as queries.

    Examples
    --------

    >>> from functools import partial
    >>> def top_customers(model, limit=10):
    ...     return model.query(...)
    >>> queries = QueryCatalog(partial(top_customers, jaffle_model))
    """

    def __init__(self, *query_funcs: Callable):
        self._queries: Dict[str, Dict[str, Any]] = {}

        for query_func in query_funcs:
            try:
                # Extract full metadata including source code
                metadata = _get_query_metadata(query_func)
                self._queries[metadata["id"]] = metadata
            except (ValueError, TypeError, AttributeError) as e:
                # If full metadata extraction fails, register with minimal info
                # This ensures queries remain executable even in restricted environments
                warnings.warn(
                    f"Could not extract full metadata for query function "
                    f"{getattr(query_func, '__name__', 'unknown')}: {e}. "
                    f"Registering with minimal metadata.",
                    UserWarning
                )
                fallback_metadata = _get_fallback_metadata(query_func, str(e))
                self._queries[fallback_metadata["id"]] = fallback_metadata

    @include_in_docs
    def get_available_queries(self) -> dict:
        """Return metadata for all registered queries.

        Returns
        -------
        dict
            Dict with ``"queries"`` and ``"query_instructions"`` keys.
        """
        # Filter out 'f' (non-serializable function reference) and 'code'
        # (verbose implementation detail unnecessary for discovery)
        queries_metadata = {
            query_id: {key: value for key, value in metadata.items() if key not in ("f", "code")}
            for query_id, metadata in self._queries.items()
        }

        return {
            "queries": queries_metadata,
            "query_instructions": (
                "IMPORTANT: Only these specific pre-defined queries are available. "
                "You CANNOT create custom queries or answer arbitrary analytical questions. "
                "For questions matching these pre-defined queries, provide a JSON payload with 'id' and 'args'. "
                "For other analytical questions, clearly explain you can only run these specific pre-defined queries."
            ),
        }

    @include_in_docs
    def explain(self, query_id: str) -> Optional[str]:
        """Return source code for a registered query by ID.

        Parameters
        ----------
        query_id : str
            Unique identifier for the query to explain.

        Returns
        -------
        str or None
            Source code string, or ``None`` if the query ID is not found.
        """
        if query_id not in self._queries:
            return None
        if "code" not in self._queries[query_id]:
            return None
        return self._queries[query_id]["code"]

    @include_in_docs
    def execute(self, query_id: str, args: dict) -> Union[rai.Fragment, pandas.DataFrame]:
        """Execute a registered query by ID.

        Parameters
        ----------
        query_id : str
            ID of the query to execute.
        args : dict
            Arguments to pass to the query function.

        Returns
        -------
        rai.Fragment or pandas.DataFrame
            Query results.

        Raises
        ------
        KeyError
            If ``query_id`` is not registered.
        """
        if query_id not in self._queries:
            available = list(self._queries.keys())
            raise KeyError(
                f"Query '{query_id}' not found. "
                f"This model has a limited catalog with only these pre-defined queries: {available}. "
                f"Custom queries cannot be executed - only these specific analytical queries are available."
            )

        query_func = self._queries[query_id]["f"]
        return query_func(**args)


def _get_query_metadata(func: Callable) -> Dict[str, Any]:
    """Extract complete metadata from a query function.

    Parameters
    ----------
    func : Callable
        The query function to extract metadata from.

    Returns
    -------
    dict
        Dict containing ``'id'``, ``'description'``, ``'args'``, ``'code'``,
        and ``'f'`` (function reference).

    Raises
    ------
    ValueError
        If function name cannot be determined.
    TypeError
        If signature cannot be extracted.
    """
    # Get function signature
    sig = inspect.signature(func)

    # Get function name
    if not hasattr(func, '__name__'):
        raise ValueError(f"Function {func} does not have a __name__ attribute")
    name = func.__name__

    # Get docstring as description
    description = inspect.getdoc(func) or ""

    # Extract arguments and their types
    args = {}
    for param_name, param in sig.parameters.items():
        if param.annotation != inspect.Parameter.empty:
            # Get string representation of the type annotation
            if hasattr(param.annotation, '__name__'):
                args[param_name] = param.annotation.__name__
            else:
                args[param_name] = str(param.annotation)
        else:
            args[param_name] = "Any"

    # Get source code - handle bound methods and other edge cases
    try:
        # For bound methods, get the underlying function
        if hasattr(func, '__func__'):
            code = inspect.getsource(getattr(func, '__func__'))
        else:
            code = inspect.getsource(func)
    except (OSError, TypeError, AttributeError):
        # If we can't get source, provide a placeholder
        code = f"# Source code for {name} (unable to retrieve in this environment)\n"

    return {
        "id": name,
        "description": description,
        "args": args,
        "code": code,
        "f": func,  # Keep function reference for execution
    }


def _get_fallback_metadata(func: Callable, error_msg: str) -> Dict[str, Any]:
    """Create minimal metadata when full extraction fails.

    Parameters
    ----------
    func : Callable
        The query function.
    error_msg : str
        Error message from the failed extraction.

    Returns
    -------
    dict
        Minimal metadata that still allows query execution.
    """
    # Try to get function name, fallback to 'unknown'
    func_name = getattr(func, '__name__', 'unknown')

    # Try to get docstring even if other metadata extraction failed
    try:
        description = inspect.getdoc(func) or ""
    except Exception:
        description = ""

    # Try to extract arguments even if source code failed
    args = {}
    try:
        sig = inspect.signature(func)
        for param_name, param in sig.parameters.items():
            if param.annotation != inspect.Parameter.empty:
                if hasattr(param.annotation, '__name__'):
                    args[param_name] = param.annotation.__name__
                else:
                    args[param_name] = str(param.annotation)
            else:
                args[param_name] = "Any"
    except Exception:
        # If signature extraction also fails, leave args empty
        pass

    return {
        "id": func_name,
        "description": description,
        "args": args,
        "code": f"# Query {func_name} (metadata extraction failed: {error_msg})\n",
        "f": func,  # Critical: keep function reference for execution
    }



from __future__ import annotations

from typing import TYPE_CHECKING

from relationalai import __version__

if TYPE_CHECKING:
    from relationalai.config.config import Config


def mk_da_headers(cfg: "Config", *, gi_setup_skipped: bool) -> dict[str, str]:
    """Build stable Direct Access headers shared across SDK layers.

    Keep this consistent between:
    - high-level exec orchestration
    - jobs/transactions gateways
    """
    return {
        "Content-Type": "application/json",
        "gi_setup_skipped": str(gi_setup_skipped),
        "X-Enable-Guard-Rails": str(cfg.jobs.enable_guard_rails),
        "user-agent": f"pyrel/{__version__}",
        "Accept": "application/json",
    }


__all__ = ["mk_da_headers"]




"""
The ``cortex_tool_resources`` module provides low-level management of Snowflake
resources (stages, stored procedures) for Cortex agent RAI tools.
"""
import json
import logging
from dataclasses import dataclass
from typing import Callable, List, Tuple, Optional, NamedTuple, Union, cast

import pandas
from snowflake import snowpark
from snowflake.snowpark.types import StringType, IntegerType, MapType, VariantType, DataType

from relationalai.util.error import exc, warn
from relationalai.util.tracing import span
from .api.agent import CortexAgentApi
from .deployment_config import DeploymentConfig
from .tool import ToolRegistry, Capability


class SprocSignature(NamedTuple):
    """Stored procedure signature for registration and cleanup."""
    name: str
    input_types: List[DataType]
    return_type: DataType

    def signature_string(self) -> str:
        """Generate SQL signature string for DROP PROCEDURE."""
        if not self.input_types:
            return f"{self.name}()"
        type_names = []
        for t in self.input_types:
            if isinstance(t, StringType):
                type_names.append("VARCHAR")
            elif isinstance(t, IntegerType):
                type_names.append("INTEGER")
            else:
                type_names.append(str(t))
        return f"{self.name}({', '.join(type_names)})"


@dataclass
class ToolResourceStatus:
    """Status information of Snowflake resources (stage, sprocs) for agent tools.

    Attributes
    ----------
    stage_exists : bool
        ``True`` if the Snowflake stage exists.
    sprocs : dict
        Dictionary mapping sproc names to existence booleans.
    """
    stage_exists: bool
    sprocs: dict  # sproc_name -> exists

    @property
    def all_exist(self) -> bool:
        """``True`` if stage and all sprocs exist."""
        return self.stage_exists and all(self.sprocs.values())

    @property
    def any_exist(self) -> bool:
        """``True`` if any resources exist."""
        return self.stage_exists or any(self.sprocs.values())


class CortexToolResources:
    """Low-level manager for the Snowflake resources behind a Cortex agent's RAI tools.

    Creates and tears down the stage and four stored procedures
    (``RAI_DISCOVER_MODELS``, ``RAI_VERBALIZE_MODEL``, ``RAI_EXPLAIN_CONCEPT``,
    ``RAI_QUERY_MODEL``) that power the agent. Sprocs are registered with
    CALLER'S RIGHTS, so SI users execute them under their own role.

    Most deployers should use ``CortexAgentManager`` instead -- it wraps this
    class and also handles the Cortex agent lifecycle. Use this class directly
    only when you need fine-grained control over resource creation.

    Parameters
    ----------
    session : snowpark.Session
        Snowpark session with database and schema already set.
    stage_name : str
        Stage for sproc code storage. Default: ``"rai_sprocs"``.
    manage_stage : bool
        Create/drop the stage automatically. Default: ``True``.
    """

    # Sproc name constants
    RAI_DISCOVER_MODELS = "RAI_DISCOVER_MODELS"
    RAI_VERBALIZE_MODEL = "RAI_VERBALIZE_MODEL"
    RAI_EXPLAIN_CONCEPT = "RAI_EXPLAIN_CONCEPT"
    RAI_QUERY_MODEL = "RAI_QUERY_MODEL"

    # Sproc signatures (shared between create and cleanup)
    # Dictionary keyed by sproc name constants for easy lookup
    SPROC_SIGNATURES = {
        RAI_DISCOVER_MODELS: SprocSignature(RAI_DISCOVER_MODELS, [], MapType()),
        RAI_VERBALIZE_MODEL: SprocSignature(RAI_VERBALIZE_MODEL, [IntegerType()], StringType()),
        RAI_EXPLAIN_CONCEPT: SprocSignature(RAI_EXPLAIN_CONCEPT, [IntegerType(), StringType()], StringType()),
        RAI_QUERY_MODEL: SprocSignature(RAI_QUERY_MODEL, [IntegerType(), StringType()], VariantType()),
    }

    def __init__(
        self,
        session: snowpark.Session,
        stage_name: str = "rai_sprocs",
        manage_stage: bool = True,
    ):
        self._session = session
        self._stage_name = stage_name
        self._manage_stage = manage_stage

        # Extract current database and schema from session
        self._database = session.get_current_database()
        self._schema = session.get_current_schema()

    def create(
        self,
        init_tools: Callable,  # Callable[[Model], ToolRegistry]
        imports: List[Tuple[str, str]],
        config: DeploymentConfig,
        agent_spec: Optional[CortexAgentApi.AgentSpec] = None,
        extra_packages: Optional[List[str]] = None,
    ) -> CortexAgentApi.AgentSpec:
        """Register RAI tool sprocs and return an agent spec with tool configuration.

        Creates the stage (if ``manage_stage``), registers all four RAI sprocs,
        and merges tool definitions into the agent spec.

        Parameters
        ----------
        init_tools : Callable
            Callable accepting a ``Model`` and returning a ``ToolRegistry``.
            The framework creates the ``Model`` (with the sproc's session) and
            passes it in. The callable should initialize concepts/properties
            on the model and return a configured ``ToolRegistry``.
        imports : list of tuple of (str, str)
            ``(source_path, module_name)`` tuples for sproc dependencies.
        config : DeploymentConfig
            Configuration with model_name, warehouse, timeouts, integrations,
            and preview settings.
        agent_spec : CortexAgentApi.AgentSpec, optional
            Base spec to merge RAI tools into. Default: new empty spec.
        extra_packages : list of str, optional
            Additional PyPI packages for sprocs.

        Returns
        -------
        CortexAgentApi.AgentSpec
            Agent spec with RAI tools configured.

        Raises
        ------
        RAIException
            If Preview capabilities are detected without
            ``allow_preview=True``, or if stage creation fails.
        """
        with span("cortex.resources.create"):
            model_name = config.model_name if config.model_name else config.agent_name
            warehouse = config.warehouse
            query_timeout_s = config.query_timeout_s
            external_access_integration = config.external_access_integration
            artifact_repository = config.artifact_repository
            allow_preview = config.allow_preview

            if warehouse:
                if isinstance(warehouse, str):
                    warehouse = warehouse.upper()
                else:
                    exc(
                        "InvalidWarehouse",
                        f"warehouse must be a string if given, got {type(warehouse)}.",
                    )

            # Validate init_tools callable and result
            if not callable(init_tools):
                exc(
                    "InvalidInitTools",
                    f"init_tools must be callable, got {type(init_tools)}.",
                )

            try:
                from relationalai.semantics import Model
                validation_model = Model.all_models[0] if len(Model.all_models) == 1 else Model(model_name)
                registry = init_tools(validation_model)
            except Exception as e:
                exc("InitToolsException",
                    f"init_tools() raised an exception during verification: {e}\n"
                    f"Please ensure the function returns a valid ToolRegistry.")

            if not isinstance(registry, ToolRegistry):
                exc(
                    "InvalidInitTools",
                    f"init_tools() must return a ToolRegistry instance, got {type(registry)}.",
                )

            if not registry.has_tools():
                exc(
                    "NoToolsRegistered",
                    "init_tools() returned ToolRegistry with no tools registered.",
                    [
                        "Ensure your init_tools function registers a tool via ToolRegistry.add().\n\n"
                        "Example:\n"
                        "  def init_tools(model):\n"
                        "      return ToolRegistry().add(\n"
                        "          model=model,\n"
                        "          description='...'\n"
                        "      )"
                    ],
                )

            # Check for Preview capabilities before deployment
            preview_capabilities = self._check_preview_capabilities_cached(registry)
            if preview_capabilities and not allow_preview:
                cap_list = ", ".join(preview_capabilities)
                exc(
                    "PreviewCapabilities",
                    f"Preview capabilities detected in deployment: {cap_list}",
                    [
                        "Preview capabilities are not Generally Available and may have limitations.\n"
                        "To deploy Preview capabilities, set allow_preview=True in your DeploymentConfig.\n\n"
                        "Example:\n"
                        "  config = DeploymentConfig(\n"
                        "      ...,\n"
                        "      allow_preview=True  # Acknowledge Preview capabilities\n"
                        "  )"
                    ],
                )

            if preview_capabilities and allow_preview:
                cap_list = ", ".join(preview_capabilities)
                warn("PreviewCapabilities",
                     f"Deploying with Preview capabilities: {cap_list}. "
                     "These may change in future releases.")

            # Create or replace stage
            if self._manage_stage:
                with span("cortex.resources.create_stage"):
                    try:
                        self._session.sql(f"CREATE OR REPLACE STAGE {self._stage_name}").collect()
                    except Exception as e:
                        exc(
                            "StageCreationFailure",
                            f"Failed to create stage '{self._stage_name}': {e}",
                            [
                                f"Ensure you have CREATE STAGE privilege on schema {self._schema}. "
                                f"Or, you can rely on an existing stage by setting manage_stage=False."
                            ]
                        )

            # Get fully qualified schema for tool resources
            schema_fqn = self._session.get_fully_qualified_current_schema()

            # Helper to register sprocs with CALLER'S RIGHTS execution
            # This means sprocs execute with the warehouse and role of the calling session,
            # not the warehouse/role used during sproc creation
            def register_sproc(func: Callable, name: str, input_types, return_type):
                with span(f"cortex.resources.register_sproc.{name}"):
                    # Suppress Snowpark's "artifact_repository ... is experimental"
                    # log warning. Snowpark emits this via the snowflake.snowpark
                    # logger whenever artifact_repository is provided.
                    snowpark_logger = logging.getLogger("snowflake.snowpark")
                    previous_level = snowpark_logger.level
                    snowpark_logger.setLevel(logging.ERROR)
                    try:
                        import relationalai
                        self._session.sproc.register(
                            func,
                            name=name,
                            input_types=input_types,
                            return_type=return_type,
                            stage_location=self._stage_name,
                            is_permanent=True,
                            replace=True,
                            packages=[
                                "snowflake-snowpark-python",
                                "snowflake-telemetry-python",
                                f"relationalai=={relationalai.__version__}",
                                *(extra_packages if extra_packages else []),
                            ],
                            imports=cast(List[Union[str, Tuple[str, str]]], imports),
                            execute_as="caller",
                            artifact_repository=artifact_repository,
                            external_access_integrations=[external_access_integration],
                        )
                    except Exception as e:
                        exc(
                            "SprocRegistrationFailure",
                            f"Failed to register stored procedure '{name}': {e}",
                            [
                                f"Ensure your role has the following privileges:\n"
                                f"  GRANT CREATE PROCEDURE ON SCHEMA {self._schema} TO ROLE <your_role>;\n"
                                f"  GRANT DATABASE ROLE snowflake.pypi_repository_user TO ROLE <your_role>;\n\n"
                                f"Also check that the stage '{self._stage_name}' exists and is accessible."
                            ]
                        )
                    finally:
                        snowpark_logger.setLevel(previous_level)

            # Initialize or merge agent spec
            if agent_spec is None:
                agent_spec = CortexAgentApi.AgentSpec()

            agent_spec = agent_spec.merge(CortexAgentApi.AgentSpec(
                instructions=CortexAgentApi.Instructions(
                    response=self._build_response_instructions(registry),
                    orchestration=self._build_orchestration_instructions(registry),
                ),
                tools=[],
                tool_resources={},
            ))

            # Ensure tools and tool_resources are initialized after merge
            assert agent_spec.tools is not None, "tools should be initialized after merge"
            assert agent_spec.tool_resources is not None, "tool_resources should be initialized after merge"

            # Register RAI_DISCOVER_MODELS
            def rai_discover_models(session: snowpark.Session):
                try:
                    from relationalai.semantics import Model
                    model = Model(model_name)
                    registry = init_tools(model)
                    result = registry.discover()

                    if isinstance(result, dict):
                        has_queries = registry.has_capability(Capability.QUERY)
                        if not has_queries:
                            result['_note'] = (
                                'IMPORTANT: These RelationalAI (RAI) models support structure exploration only - data queries are not available. '
                                'RAI tools help users understand model structure and semantics. '
                                'Guide users to ask questions like: "What is the Customer concept?", "Explain the relationship between Orders and Customers", '
                                '"What are the key concepts in this model?", or "Show me the structure of the Order concept". '
                                'Make it clear these are RAI-powered capabilities focused on semantic understanding. '
                                'DO NOT suggest analytical questions like "show me top customers" or "calculate total revenue".'
                            )
                        else:
                            result['_note'] = (
                                'IMPORTANT: This model has a PREVIEW query capability with a limited catalog of specific pre-defined queries. '
                                'You can ONLY execute the exact queries listed in the model metadata. '
                                'For any analytical question, first check if it matches one of the available pre-defined queries. '
                                'If there is no matching pre-defined query, you CANNOT answer the question with data - '
                                'instead, guide users to understand the data model structure using RAI_VERBALIZE_MODEL and RAI_EXPLAIN_CONCEPT. '
                                'Frame this as: "This model provides specific curated analytical queries" - these are intentional limitations, not bugs.'
                            )

                    return result
                except Exception as e:
                    import traceback
                    error_msg = f"Error in rai_discover_models: {str(e)}\n{traceback.format_exc()}"
                    print(error_msg)
                    return {"error": error_msg, "models": []}

            sig = self.SPROC_SIGNATURES[self.RAI_DISCOVER_MODELS]
            register_sproc(rai_discover_models, sig.name, sig.input_types, sig.return_type)
            agent_spec.tools = [t for t in agent_spec.tools if t.tool_spec.name != sig.name]
            agent_spec.tools.append(CortexAgentApi.Tool(
                CortexAgentApi.ToolSpec(
                    type="generic",
                    name=sig.name,
                    description="Discover available data models in the RelationalAI registry. Returns models that can be explored to answer business questions.",
                    input_schema=None,
                )
            ))
            agent_spec.tool_resources[sig.name] = CortexAgentApi.ToolResourceSpec(
                execution_environment=CortexAgentApi.ExecutionEnvironment(
                    query_timeout=query_timeout_s,
                    type="warehouse",
                    warehouse=warehouse
                ),
                identifier=f"{schema_fqn}.{sig.name}",
                name=f"{sig.name}()",
                type="procedure",
            )

            # Register RAI_VERBALIZE_MODEL
            def rai_verbalize_model(session: snowpark.Session, param1: int):
                try:
                    from relationalai.semantics import Model
                    model = Model(model_name)
                    return init_tools(model).verbalize(param1)
                except Exception as e:
                    import traceback
                    error_msg = f"Error in rai_verbalize_model for model '{param1}': {str(e)}\n{traceback.format_exc()}"
                    print(error_msg)
                    return error_msg

            sig = self.SPROC_SIGNATURES[self.RAI_VERBALIZE_MODEL]
            register_sproc(rai_verbalize_model, sig.name, sig.input_types, sig.return_type)
            agent_spec.tools = [t for t in agent_spec.tools if t.tool_spec.name != sig.name]
            agent_spec.tools.append(CortexAgentApi.Tool(
                CortexAgentApi.ToolSpec(
                    type="generic",
                    name=sig.name,
                    description="Get detailed natural language descriptions of a data model from the RelationalAI registry. Use this RAI tool to understand model structure and present business-friendly explanations to users.",
                    input_schema={
                        "type": "object",
                        "properties": {
                            "arg1": {
                                "description": f"Numeric identifier obtained from {self.RAI_DISCOVER_MODELS}",
                                "type": "integer"
                            }
                        },
                        "required": ["arg1"]
                    },
                )
            ))
            agent_spec.tool_resources[sig.name] = CortexAgentApi.ToolResourceSpec(
                execution_environment=CortexAgentApi.ExecutionEnvironment(
                    query_timeout=query_timeout_s,
                    type="warehouse",
                    warehouse=warehouse
                ),
                identifier=f"{schema_fqn}.{sig.name}",
                name=f"{sig.name}(INTEGER)",
                type="procedure",
            )

            # Register RAI_EXPLAIN_CONCEPT
            def rai_explain_concept(session: snowpark.Session, param1: int, param2: str):
                try:
                    from relationalai.semantics import Model
                    model = Model(model_name)
                    return init_tools(model).explain(param1, param2)
                except Exception as e:
                    import traceback
                    error_msg = f"Error in rai_explain_concept for model '{param1}', concept '{param2}': {str(e)}\n{traceback.format_exc()}"
                    print(error_msg)
                    return error_msg

            sig = self.SPROC_SIGNATURES[self.RAI_EXPLAIN_CONCEPT]
            register_sproc(rai_explain_concept, sig.name, sig.input_types, sig.return_type)
            agent_spec.tools = [t for t in agent_spec.tools if t.tool_spec.name != sig.name]
            agent_spec.tools.append(CortexAgentApi.Tool(
                CortexAgentApi.ToolSpec(
                    type="generic",
                    name=sig.name,
                    description="Get explanations for specific concepts or pre-defined queries in a RelationalAI model. Pass a concept name or a query ID — both obtained from RAI_DISCOVER_MODELS — to receive the appropriate explanation.",
                    input_schema={
                        "type": "object",
                        "properties": {
                            "arg1": {
                                "description": f"Numeric model_id acquired from {self.RAI_DISCOVER_MODELS}",
                                "type": "integer"
                            },
                            "arg2": {
                                "description": f"A concept name or query ID, both obtained from {self.RAI_DISCOVER_MODELS}",
                                "type": "string"
                            }
                        },
                        "required": ["arg1", "arg2"]
                    },
                )
            ))
            agent_spec.tool_resources[sig.name] = CortexAgentApi.ToolResourceSpec(
                execution_environment=CortexAgentApi.ExecutionEnvironment(
                    query_timeout=query_timeout_s,
                    type="warehouse",
                    warehouse=warehouse
                ),
                identifier=f"{schema_fqn}.{sig.name}",
                name=f"{sig.name}(INTEGER, VARCHAR)",
                type="procedure",
            )

            # Register RAI_QUERY_MODEL
            def rai_query_model(session: snowpark.Session, param1: int, param2: str):
                import traceback
                try:
                    from relationalai.semantics import Model
                    model = Model(model_name)
                    registry = init_tools(model)
                except Exception as init_err:
                    error_trace = traceback.format_exc()
                    print(f"Error in rai_query_model: {str(init_err)}\n{error_trace}")
                    return [{"error": f"Query execution failed: {str(init_err)}", "model_id": param1,
                             "body_preview": param2[:100] if param2 else None, "traceback": error_trace[:500]}]

                model_id = param1
                body = param2

                if not registry.tool_has_capability(model_id, Capability.QUERY):
                    return [{
                        "error": "This model does not have query execution capability.",
                        "model_id": model_id,
                        "available_capabilities": ["discover", "verbalize", "explain"],
                        "guidance": "You can help users understand the data model structure and semantics, but cannot execute data queries."
                    }]

                try:
                    query_dict = json.loads(body)
                except json.JSONDecodeError as json_err:
                    return [{"error": f"Invalid JSON in body parameter: {str(json_err)}", "body_received": body[:100],
                             "expected_format": '{"id": "<query_id>", "args": {<params>}}'}]

                if not isinstance(query_dict, dict) or "id" not in query_dict:
                    # Retrieve available queries to guide the agent
                    available_queries = []
                    try:
                        tool = registry._tools[model_id]
                        query_meta = tool.queries()
                        if "queries" in query_meta:
                            for qid, qinfo in query_meta["queries"].items():
                                entry = {"id": qid}
                                if "args" in qinfo:
                                    entry["args"] = qinfo["args"]
                                if "description" in qinfo:
                                    entry["description"] = qinfo["description"]
                                available_queries.append(entry)
                    except Exception:
                        pass
                    return [{
                        "error": "Invalid query payload format.",
                        "expected_format": '{"id": "<query_id>", "args": {<params>}}',
                        "body_received": body[:200] if body else None,
                        "available_queries": available_queries,
                        "guidance": "The payload must be a JSON object with an 'id' key matching one of the available query IDs, and an 'args' key with the query parameters."
                    }]

                try:
                    result = registry.query(model_id, query_dict)
                except KeyError as key_err:
                    # Could be model not found or query ID not found — include available queries
                    available_queries = []
                    try:
                        tool = registry._tools[model_id]
                        query_meta = tool.queries()
                        if "queries" in query_meta:
                            for qid, qinfo in query_meta["queries"].items():
                                entry = {"id": qid}
                                if "args" in qinfo:
                                    entry["args"] = qinfo["args"]
                                if "description" in qinfo:
                                    entry["description"] = qinfo["description"]
                                available_queries.append(entry)
                    except (IndexError, KeyError):
                        try:
                            available_models = list(range(len(registry._tools)))
                        except Exception:
                            available_models = ["Unable to retrieve available models"]
                        return [{"error": f"Model '{model_id}' not found. Available models: {available_models}"}]
                    return [{
                        "error": f"Query not found: {str(key_err)}",
                        "model_id": model_id,
                        "available_queries": available_queries,
                        "expected_format": '{"id": "<query_id>", "args": {<params>}}'
                    }]
                except Exception as exec_err:
                    error_trace = traceback.format_exc()
                    print(f"Error in rai_query_model: {str(exec_err)}\n{error_trace}")
                    return [{"error": f"Query execution failed: {str(exec_err)}", "model_id": model_id,
                             "body_preview": body[:100] if body else None, "traceback": error_trace[:500]}]

                if isinstance(result, pandas.DataFrame):
                    return result.to_dict(orient="records")
                return result

            sig = self.SPROC_SIGNATURES[self.RAI_QUERY_MODEL]
            register_sproc(rai_query_model, sig.name, sig.input_types, sig.return_type)
            agent_spec.tools = [t for t in agent_spec.tools if t.tool_spec.name != sig.name]

            query_description = "Execute a pre-defined RAI query from the model's query catalog. ONLY call this for questions that match specific pre-defined queries discovered from RAI_DISCOVER_MODELS."
            if Capability.QUERY.is_preview:
                query_description += " [PREVIEW - Limited catalog of curated queries. Cannot create custom queries.]"

            if not registry.has_capability(Capability.QUERY):
                query_description += " [NO MODELS WITH QUERY CAPABILITY - This tool will return an error if called]"

            agent_spec.tools.append(CortexAgentApi.Tool(
                CortexAgentApi.ToolSpec(
                    type="generic",
                    name=sig.name,
                    description=query_description,
                    input_schema={
                        "type": "object",
                        "properties": {
                            "arg1": {
                                "description": f"Numeric model_id obtained from {self.RAI_DISCOVER_MODELS}",
                                "type": "integer"
                            },
                            "arg2": {
                                "description": (
                                    f'JSON string with the query payload. '
                                    f'Must contain "id" (the query ID from {self.RAI_DISCOVER_MODELS}) '
                                    f'and "args" (an object mapping parameter names to values). '
                                    f'Example: {{"id": "my_query", "args": {{"limit": 10}}}}. '
                                    f'For queries with no parameters, use empty args: {{"id": "my_query", "args": {{}}}}.'
                                ),
                                "type": "string"
                            }
                        },
                        "required": ["arg1", "arg2"]
                    },
                )
            ))
            agent_spec.tool_resources[sig.name] = CortexAgentApi.ToolResourceSpec(
                execution_environment=CortexAgentApi.ExecutionEnvironment(
                    query_timeout=query_timeout_s,
                    type="warehouse",
                    warehouse=warehouse
                ),
                identifier=f"{schema_fqn}.{sig.name}",
                name=f"{sig.name}(INTEGER, VARCHAR)",
                type="procedure",
            )

            return agent_spec

    def _check_preview_capabilities_cached(self, registry: ToolRegistry) -> List[str]:
        """Check if any Preview capabilities are configured in the tool registry."""
        preview_caps = []

        try:
            for capability in Capability:
                if capability.is_preview and registry.has_capability(capability):
                    preview_caps.append(capability.value)
        except Exception as e:
            warn("PreviewCheck", f"Could not check for Preview capabilities: {e}")

        return preview_caps

    def _build_response_instructions(self, registry: ToolRegistry) -> str:
        """Build response instructions based on available capabilities."""
        base = [
            "You are a professional assistant that uses tools to answer business questions from non-technical users.\n"
            "Your response must be based off the results of some tool. If none of the available tools can be used to answer a question, "
            "your response should indicate that and nothing else. Do not present general, non-tool based knowledge in the response. "
            "Focus more on clearly displaying the results, rather than providing exhaustive analysis. "
            "The user should be able to understand the results and ask follow-up questions.\n\n"
            "CRITICAL PRESENTATION RULES:\n"
            "- NEVER show Python code, SQL DDL, or any programming syntax to end users\n"
            "- NEVER display code snippets, function definitions, or implementation details\n"
            "- Translate all technical content to natural language descriptions that business users can understand\n"
            "- Express formulas using mathematical notation (e.g., \u00d7 for multiplication, \u00f7 for division) NOT Python operators (*, /)\n"
            "- Focus on WHAT the business logic does, not HOW it is implemented in code\n"
            "- Describe relationships, rules, and calculations in plain business language\n"
            "- If tool output contains code, summarize its meaning in natural language instead of showing the code\n\n"
            "BRANDING REQUIREMENTS:\n"
            "- When introducing capabilities, mention they are 'powered by RelationalAI' or reference 'RAI tools'\n"
            "- When explaining what models are available, note they are from 'the RelationalAI registry'\n"
            "- In refusal explanations, frame limitations as 'RAI tools focus on...' rather than generic 'available tools'\n"
            "- Natural mentions preferred: integrate 'RelationalAI' or 'RAI' into explanations organically\n\n"
        ]

        if registry.has_capability(Capability.QUERY):
            base.append(
                "QUERIES CAPABILITY - CRITICAL COMMUNICATION RULES:\n"
                "- When introducing query capabilities, ALWAYS clarify these are 'specific pre-defined queries' not arbitrary query execution\n"
                "- When listing available queries, frame them as 'These specific analytical queries are available' NOT as examples\n"
                "- For questions that don't match pre-defined queries, explain: 'This model has a limited set of pre-defined queries. Your question doesn't match any available queries.'\n"
                "- NEVER suggest you can answer arbitrary analytical questions - only the pre-defined queries listed in the catalog\n"
                "- Frame as intentional design: 'This model provides specific curated analytical queries' NOT as a limitation or gap\n\n"
            )

        if not registry.has_capability(Capability.QUERY):
            base.append(
                "IMPORTANT - When responding to onboarding prompts (like 'what can I ask about?' or 'help me get started'):\n"
                "- DO NOT suggest analytical queries like 'show me top customers' or 'calculate revenue' - you cannot answer these\n"
                "- ONLY guide users to ask questions about model STRUCTURE and SEMANTICS\n"
                "- Emphasize that you can explain relationships, definitions, and business rules, but NOT execute data queries\n"
                "- Make it crystal clear that data analysis and aggregation queries are not available\n\n"
            )

        return "".join(base)

    def _build_orchestration_instructions(self, registry: ToolRegistry) -> str:
        """Build orchestration instructions based on available capabilities."""
        base = ("Use RAI tools to understand data models, add context to answers, "
                "and ensure your answers are semantically correct. "
                "When presenting information to users, focus on semantic meaning for business users and avoid showing code snippets. "
                "When explaining capabilities to users, naturally mention that these are 'RAI tools' or 'RelationalAI-powered capabilities'. "
                "Frame model discovery as accessing 'the RelationalAI registry' or 'RAI models'. ")

        instructions = [base]

        if not registry.has_capability(Capability.QUERY):
            instructions.append(
                "CRITICAL: Query capability is NOT available for any models in the registry. "
                "Do NOT attempt to call RAI_QUERY_MODEL under any circumstances. "
                "You can ONLY discover models, verbalize their structure, and explain concepts. "
                "You CANNOT execute data queries, aggregations, or analytical operations. "
                "When explaining these limitations, position your capabilities as: 'RAI tools focus on helping you understand data model structure and semantics.' "
            )

        instructions.append(
            "Always discover RAI models first when beginning a conversation so you understand what you can answer questions about.\n\n"
        )

        if registry.has_capability(Capability.VERBALIZE):
            instructions.append(
                "Use RAI_VERBALIZE_MODEL to get detailed descriptions of data model structure, relationships, and semantics. "
            )

            if registry.has_capability(Capability.QUERY):
                instructions.append("This helps you understand the model's concepts and relationships.\n\n")
            else:
                instructions.append("Focus on explaining the model structure and relationships to users.\n\n")

        if registry.has_capability(Capability.EXPLAIN):
            instructions.append(
                "Use RAI_EXPLAIN_CONCEPT to understand business rules for a specific concept or to get a description of a pre-defined query. Both concept names and query IDs come from RAI_DISCOVER_MODELS.\n\n"
            )

        if registry.has_capability(Capability.QUERY):
            instructions.append(
                "Use RAI_QUERY_MODEL to execute ONLY the specific pre-defined analytical queries available in the model's query catalog. "
                "BEFORE attempting any query, check if the user's question matches one of the pre-defined queries discovered from RAI_DISCOVER_MODELS. "
                "If the question does NOT match a pre-defined query, do NOT call RAI_QUERY_MODEL. Instead, clearly explain that only specific pre-defined queries are available. "
                "Never attempt to construct custom queries - only use the exact pre-defined queries from the catalog.\n\n"
                "QUERY PAYLOAD FORMAT: The arg2 parameter must be a JSON string with this exact structure: "
                '{\"id\": \"<query_id>\", \"args\": {<params>}}. '
                "The 'id' must match a query ID from RAI_DISCOVER_MODELS. "
                "The 'args' must be an object mapping parameter names to values as described in the query metadata. "
                "For queries with no parameters, use empty args: "
                '{\"id\": \"<query_id>\", \"args\": {}}.\n\n'
            )

        return "".join(instructions)

    def cleanup(self, silent: bool = True) -> None:
        """Drop all RAI sprocs and the stage (if ``manage_stage``).

        Parameters
        ----------
        silent : bool
            Ignore errors for non-existent resources. Default: ``True``.

        Returns
        -------
        None
        """
        with span("cortex.resources.cleanup"):
            # Delete all sprocs with full signatures
            for sig in self.SPROC_SIGNATURES.values():
                try:
                    # Use full signature to handle overloaded procedures
                    self._session.sql(
                        f"DROP PROCEDURE IF EXISTS {self._database}.{self._schema}.{sig.signature_string()}"
                    ).collect()
                except Exception as e:
                    if not silent:
                        raise
                    warn("Cleanup", f"Failed to delete sproc {sig.name}: {e}")

            # Delete stage only if we manage it
            if self._manage_stage:
                try:
                    self._session.sql(
                        f"DROP STAGE IF EXISTS {self._database}.{self._schema}.{self._stage_name}"
                    ).collect()
                except Exception as e:
                    if not silent:
                        raise
                    warn("Cleanup", f"Failed to delete stage {self._stage_name}: {e}")

    def exists(self) -> ToolResourceStatus:
        """Check existence of the stage and all RAI sprocs.

        Returns
        -------
        ToolResourceStatus
            Status of each Snowflake resource.
        """
        # Check stage existence
        stage_exists = False
        try:
            result = self._session.sql(
                f"SHOW STAGES LIKE '{self._stage_name}' IN {self._database}.{self._schema}"
            ).collect()
            stage_exists = len(result) > 0
        except Exception:
            pass

        # Check sproc existence
        sproc_status = {}
        for sproc_name in self.SPROC_SIGNATURES.keys():
            exists = False
            try:
                result = self._session.sql(
                    f"SHOW PROCEDURES LIKE '{sproc_name}' IN {self._database}.{self._schema}"
                ).collect()
                exists = len(result) > 0
            except Exception:
                pass
            sproc_status[sproc_name] = exists

        return ToolResourceStatus(stage_exists=stage_exists, sprocs=sproc_status)
